"""
Default Exp form the YOLOX inference.
"""
import os

from yolox.exp import Exp as MyExp


class Exp(MyExp):
    """
    Experiment class for YOLOX model.

    Copied from "https://github.com/DanceTrack/DanceTrack/blob/main/ByteTrack/exps/example/dancetrack/yolox_x.py"
    """
    def __init__(self):
        super().__init__()
        self.num_classes = 1
        self.class_names = ('pedestrian',)
        self.depth = 1.33
        self.width = 1.25
        self.exp_name = os.path.split(os.path.realpath(__file__))[1].split('.')[0]
        self.train_ann = 'train.json'
        self.val_ann = 'val.json'
        self.test_ann = 'test.json'

        self.input_size = (800, 1440)
        self.test_size = (800, 1440)
        self.random_size = (18, 32)
        self.max_epoch = 8
        self.print_interval = 20
        self.eval_interval = 5
        self.test_conf = 0.1
        self.nmsthre = 0.7
        self.no_aug_epochs = 1
        self.basic_lr_per_img = 0.001 / 64.0
        self.warmup_epochs = 1


DEFAULT_EXP_PATH = __file__
DEFAULT_EXP_NAME = Exp.__name__



"""
Object detection factory method.
Use `OBJECT_DETECTION_CATALOG.register` to extend supported object detection algorithms.
"""
from typing import Optional

from motrack.object_detection.algorithms.base import ObjectDetectionInference
# noinspection PyUnresolvedReferences
from motrack.object_detection.algorithms.yolov8 import YOLOv8Inference  # pylint: disable=unused-import
# noinspection PyUnresolvedReferences
from motrack.object_detection.algorithms.yolox import YOLOXInference  # pylint: disable=unused-import
from motrack.object_detection.catalog import OBJECT_DETECTION_CATALOG
from motrack.utils.lookup import LookupTable


def object_detection_inference_factory(
    name: str,
    params: dict,
    lookup: Optional[LookupTable] = None
) -> ObjectDetectionInference:
    """
    Creates OD inference.

    Args:
        name: OD inference name (type)
        params: CTor parameters
        lookup: Used to decode classes

    Returns:
        Initialized OD inference
    """
    return OBJECT_DETECTION_CATALOG[name](**params, lookup=lookup)



"""
Implementation of Filter Sort tracker
"""
from typing import Tuple, Optional, List, Union

import numpy as np
from abc import ABC, abstractmethod
import copy

from motrack.filter import filter_factory
from motrack.library.cv.bbox import PredBBox, BBox
from motrack.tracker.trackers.algorithms.base import Tracker
from motrack.tracker.tracklet import Tracklet, TrackletState, TrackletCommonData
from motrack.cmc import cmc_factory
from motrack.reid import reid_inference_factory
from collections import deque


class MotionReIDBasedTracker(Tracker, ABC):
    """
    Baseline Sort tracker components:
    - ReID: HungarianIOU
    - Motion model: Filter prior
    - Combining detection and motion model: Filter posterior
    """
    def __init__(
        self,
        filter_name: str,
        filter_params: dict,
        cmc_name: Optional[str] = None,
        cmc_params: Optional[dict] = None,
        reid_name: Optional[str] = None,
        reid_params: Optional[str] = None,
        reid_detection_threshold: Optional[float] = None,
        initialization_threshold: int = 3,
        remember_threshold: int = 1,
        new_tracklet_detection_threshold: Optional[float] = None,
        use_observation_if_lost: bool = False,
        appearance_ema_momentum: float = 0.95,
        appearance_buffer: int = 0
    ):
        """
        Args:
            filter_name: Filter name
            filter_params: Filter params
            cmc_name: CMC name
            cmc_params: CMC params
            reid_name: ReID name
            reid_params: ReID params

            reid_detection_threshold: Filter bboxes that have more than `reid_detection_threshold`
                detection score before applying ReID. Mandatory for trackers that use low detection scores
                like `BYTE`.

            remember_threshold: How long does the tracklet without any detection matching
                If tracklet isn't matched for more than `remember_threshold` frames then
                it is deleted.
            initialization_threshold: Number of frames until tracklet becomes active
            new_tracklet_detection_threshold: Threshold to accept new tracklet
            use_observation_if_lost: When re-finding tracklet, use observation instead of estimation
            appearance_ema_momentum: Appearance EMA (Exponential Moving Average) momentum
            appearance_buffer: Appearance buffer length (set to a non-zero non-negative integer to activate)
        """
        super().__init__()

        self._filter = filter_factory(filter_name, filter_params)

        self._cmc = None
        if cmc_name is not None:
            cmc_params = {} if cmc_params is None else cmc_params
            self._cmc = cmc_factory(cmc_name, cmc_params)

        self._reid = None
        if reid_name is not None:
            reid_params = {} if reid_params is None else reid_params
            self._reid = reid_inference_factory(reid_name, reid_params)

        # Hyperparameters
        self._reid_detection_threshold = reid_detection_threshold
        self._new_tracklet_detection_threshold = new_tracklet_detection_threshold
        self._initialization_threshold = initialization_threshold
        self._remember_threshold = remember_threshold
        self._use_observation_if_lost = use_observation_if_lost

        # ReID hyperparameters
        self._appearance_ema_momentum = appearance_ema_momentum
        self._appearance_buffer = appearance_buffer

        # State
        self._filter_states = {}
        self._next_id = 0

    @staticmethod
    def _raw_to_bbox(tracklet: Tracklet, raw: np.ndarray, conf: Optional[float] = None) -> PredBBox:
        """
        Converts raw tensor to PredBBox for tracked object.

        Args:
            tracklet: Tracklet
            raw: raw bbox coords
            conf: BBox confidence

        Returns:
            PredBBox
        """
        bbox_raw = raw.tolist()
        return PredBBox.create(
            bbox=BBox.from_xywh(*bbox_raw, clip=False),
            label=tracklet.bbox.label,
            conf=tracklet.bbox.conf if conf is None else conf
        )

    def _initiate(self, tracklet_id: int, detection: PredBBox) -> None:
        """
        Initiates new tracking object state.

        Args:
            tracklet_id: Tracklet id
            detection: Initial object detection
        """
        measurement = detection.as_numpy_xywh(dtype=np.float32)
        state = self._filter.initiate(measurement)
        self._filter_states[tracklet_id] = state

    def _predict(self, tracklet: Tracklet) -> Tuple[PredBBox, np.ndarray, np.ndarray]:
        """
        Estimates object prior position

        Args:
            tracklet: Tracklet (required to fetch state, and bbox info)

        Returns:
            Motion model prior estimation
        """
        assert tracklet.id in self._filter_states, f'Tracklet id "{tracklet.id}" can\'t be found in filter states!'
        state = self._filter_states[tracklet.id]
        state = self._filter.predict(state)
        prior_mean, prior_std = self._filter.project(state)
        self._filter_states[tracklet.id] = state
        bbox = self._raw_to_bbox(tracklet, prior_mean)
        return bbox, prior_mean, prior_std

    def _update(self, tracklet: Tracklet, detection: PredBBox) -> Tuple[PredBBox, np.ndarray, np.ndarray]:
        """
        Estimates object posterior position based on the matched detection.

        Args:
            tracklet: Tracklet (required to fetch state, and bbox info)

        Returns:
            Motion model posterior estimation
        """
        measurement = detection.as_numpy_xywh()

        state = self._filter_states[tracklet.id]
        state = self._filter.update(state, measurement)
        posterior_mean, posterior_std = self._filter.project(state)
        self._filter_states[tracklet.id] = state
        bbox = self._raw_to_bbox(tracklet, posterior_mean, conf=detection.conf)
        return bbox, posterior_mean, posterior_std

    def _missing(self, tracklet: Tracklet) -> Tuple[PredBBox, np.ndarray, np.ndarray]:
        """
        Estimates object posterior position for unmatched tracklets (with any detections)

        Args:
            tracklet: Tracklet (required to fetch state, and bbox info)

        Returns:
            Motion model posterior (missing) estimation
        """
        state = self._filter_states[tracklet.id]
        state = self._filter.missing(state)
        posterior_mean, posterior_std = self._filter.project(state)
        self._filter_states[tracklet.id] = state
        bbox = self._raw_to_bbox(tracklet, posterior_mean)
        return bbox, posterior_mean, posterior_std

    def _delete(self, tracklet_id: int) -> None:
        """
        Deletes filter state for deleted tracklet id.

        Args:
            tracklet_id: Tracklet id
        """
        self._filter_states.pop(tracklet_id)

    def _create_new_tracklets(
        self,
        detections: List[PredBBox],
        frame_index: int,
        objects_features: Optional[np.ndarray] = None
    ) -> List[Tracklet]:
        """
        Creates new tracklets based on unmatched detections.

        Args:
            detections: List of unmatched detections
            frame_index: Current frame index
            objects_features: List of unmatched detections' object features

        Returns:
            New tracklets
        """
        new_tracklets: List[Tracklet] = []
        for d_i, detection in enumerate(detections):
            if self._new_tracklet_detection_threshold is not None and detection.conf < self._new_tracklet_detection_threshold:
                continue

            new_tracklet = Tracklet(
                bbox=copy.deepcopy(detection),
                frame_index=frame_index,
                _id=self._next_id,
                state=TrackletState.NEW if frame_index > self._initialization_threshold else TrackletState.ACTIVE
            )
            self._next_id += 1
            new_tracklets.append(new_tracklet)
            self._initiate(new_tracklet.id, detection)

            # Set initial tracklet feature if ReId model is used
            if objects_features is not None:
                new_tracklet.set(TrackletCommonData.APPEARANCE, objects_features[d_i])
                if self._appearance_buffer > 0:
                    buffer = deque(maxlen=self._appearance_buffer)
                    buffer.append(objects_features[d_i])
                    new_tracklet.set(TrackletCommonData.APPEARANCE_BUFFER, buffer)

        return new_tracklets

    def _update_tracklets(
        self,
        matched_tracklets: List[Tracklet],
        matched_detections: List[PredBBox],
        matched_object_features: Union[np.ndarray, List[Optional[np.ndarray]]],
        frame_index: int
    ) -> List[Tracklet]:
        """
        Updates tracklets with detections and object appearance features (if used).

        Args:
            matched_tracklets: List of matched tracklets
            matched_detections: List of matched detections
            frame_index: Current frame index

        Returns:
            List of updated tracklets
        """
        assert len(matched_tracklets) == len(matched_tracklets), \
            'Number of matched tracklets and number of matched detections do not match! ' \
            f'Got: {len(matched_tracklets)}, {len(matched_tracklets)}'

        for tracklet, det_bbox in zip(matched_tracklets, matched_detections):
            tracklet_bbox, _, _ = self._update(tracklet, det_bbox)
            new_bbox = det_bbox if self._use_observation_if_lost and tracklet.state != TrackletState.ACTIVE \
                else tracklet_bbox

            new_state = TrackletState.ACTIVE
            if tracklet.state == TrackletState.NEW and tracklet.total_matches + 1 < self._initialization_threshold:
                new_state = TrackletState.NEW
            tracklet.update(new_bbox, frame_index, state=new_state)

        self._update_appearances(
            tracklets=matched_tracklets,
            object_features=matched_object_features
        )

        return matched_tracklets

    def _update_appearances(self, tracklets: List[Tracklet], object_features: Union[np.ndarray, List[Optional[np.ndarray]]]) -> None:
        """
        Updates Tracklet appearance.

        Args:
            tracklets: Tracklets
            object_features: Detected objects features
        """
        if self._reid is None:
            return

        assert len(tracklets) == len(object_features), \
            'Number of matched tracklets and number of matched detection features do not match! ' \
            f'Got: {len(tracklets)}, {len(object_features)}'

        for i, tracklet in enumerate(tracklets):
            current_emb = object_features[i]
            if current_emb is None:
                continue

            emb = tracklet.get(TrackletCommonData.APPEARANCE)
            if emb is None:
                emb = current_emb
            else:
                emb: np.ndarray
                emb = self._appearance_ema_momentum * emb + (1 - self._appearance_ema_momentum) * current_emb

            emb /= np.linalg.norm(emb)
            tracklet.set(TrackletCommonData.APPEARANCE, emb)

            if self._appearance_buffer > 0:
                buffer = tracklet.get(TrackletCommonData.APPEARANCE_BUFFER)
                buffer.append(current_emb)

    def _handle_lost_tracklets(self, lost_tracklets: List[Tracklet]) -> None:
        """
        Handles lost tracklets:
        - Deletes tracklets that are lost for more than `self._remember_threshold` frames
        - Deletes new tracklets that are lost before initialized
        - Update lost tracklets position using the motion model

        Args:
            lost_tracklets: Lost tracklets that potentially should be deleted
        """
        for tracklet in lost_tracklets:
            if tracklet.lost_time > self._remember_threshold \
                    or tracklet.state == TrackletState.NEW:
                tracklet.state = TrackletState.DELETED
                self._delete(tracklet.id)
            else:
                tracklet_bbox, _, _ = self._missing(tracklet)
                tracklet.update(tracklet_bbox, tracklet.frame_index, state=TrackletState.LOST)

    def _perform_cmc(self, frame: np.ndarray, frame_index: int, bboxes: List[PredBBox]) -> List[PredBBox]:
        """
        Performs CMC on detected bounding boxes and updates motion filter states.

        Args:
            frame: Frame (image)
            frame_index: Current frame index
            bboxes: List of bounding boxes

        Returns:
            Updated bounding boxes
        """
        if self._cmc is None:
            return bboxes

        warp = self._cmc.apply(frame, frame_index, scene=self._scene)
        self._filter_states = {t_id: self._filter.affine_transform(state, warp) for t_id, state in self._filter_states.items()}

        corrected_bboxes: List[PredBBox] = []
        for bbox in bboxes:
            bbox = PredBBox.create(
                bbox=bbox.affine_transform(warp),
                label=bbox.label,
                conf=bbox.conf
            )
            corrected_bboxes.append(bbox)

        return corrected_bboxes


    def _extract_reid_features(self, frame: np.ndarray, frame_index: int, bboxes: List[PredBBox]) -> Optional[np.ndarray]:
        """
        Extracts features for each detected bounding box. Skips if ReID is not used

        Args:
            frame: Frame (image)
            frame_index: Current frame index
            bboxes: List of bounding boxes

        Returns:
            ReId features for each detected bounding box if ReId is used else None
        """
        if self._reid is None:
            # ReID is not used
            return None

        if self._reid_detection_threshold is not None:
            # Filter bboxes with high confidence
            # Remove bboxes with low confidence
            bboxes = [bbox for bbox in bboxes if bbox.conf >= self._reid_detection_threshold]

        return self._reid.extract_objects_features(frame, bboxes, frame_index=frame_index, scene=self._scene)

    def track(
        self,
        tracklets: List[Tracklet],
        detections: List[PredBBox],
        frame_index: int,
        frame: Optional[np.ndarray] = None
    ) -> List[Tracklet]:
        frame_index -= 1
        tracklets, prior_tracklet_bboxes = self._track_predict(tracklets, frame_index, frame=frame)
        objects_features = self._extract_reid_features(frame, frame_index, detections)
        return self._track(tracklets, prior_tracklet_bboxes, detections, frame_index,
                           objects_features=objects_features, frame=frame)

    def _track_predict(self, tracklets: List[Tracklet], frame_index: int, frame: Optional[np.ndarray] = None):
        """
        Performs tracker tracklet filtering motion prediction steps:
        - Removes DELETED trackelts
        - Uses filter to obtain prior tracklet position estimates
        - Optionally performs CMC for prediction corrections

        Args:
            tracklets: List of tracklets
            frame_index: Current frame index
            frame: Current frame image (optional)

        Returns:
            - Filtered tracklets
            - Tracklet object position predictions for the current frame
        """
        tracklets = [t for t in tracklets if t.state != TrackletState.DELETED]  # Remove deleted tracklets

        # Estimate priors for all tracklets
        prior_tracklet_estimates = [self._predict(t) for t in tracklets]
        prior_tracklet_bboxes = [bbox for bbox, _, _ in prior_tracklet_estimates]
        prior_tracklet_bboxes = self._perform_cmc(frame, frame_index, prior_tracklet_bboxes)

        return tracklets, prior_tracklet_bboxes

    @abstractmethod
    def _track(
        self,
        tracklets: List[Tracklet],
        prior_tracklet_bboxes: List[PredBBox],
        detections: List[PredBBox],
        frame_index: int,
        objects_features: Optional[np.ndarray] = None,
        frame: Optional[np.ndarray] = None
    ) -> List[Tracklet]:
        """
        Main tracker logic (after prior estimates and CMC).

        Args:
            tracklets: List of active trackelts
            prior_tracklet_bboxes: Tracklet prior position bboxes
            detections: Lists of new detections
            frame_index: Current frame number
            objects_features: Object appearance features
            frame: Pass frame in case CMC or appearance based association is used

        Returns:
            Tracks (active, lost, new, deleted, ...)
        """



"""
Motrack tools interface.
"""
from motrack.tools.inference import run_tracker_inference
from motrack.tools.postprocess import run_tracker_postprocess
from motrack.tools.visualize import run_visualize_tracker_inference, run_visualize_detections



"""
Camera motion compensation interface.
"""
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np


class CameraMotionCompensation(ABC):
    @abstractmethod
    def apply(self, frame: np.ndarray, frame_index: int, scene: Optional[str] = None) -> np.ndarray:
        """
        Calculates approximated affine transformations applied
        to the image between current and last frame.

        Args:
            frame: Current video frame
            frame_index: Frame index
            scene: Scene name (optional)

        Returns:
            Affine 2x3 matrix (includes translation)
        """



"""
Implementation of Move (hybrid) association method.
"""
from typing import Optional, List

import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.algorithms.compose import ComposeAssociationAlgorithm
from motrack.tracker.matching.algorithms.iou import IoUAssociation, LabelGatingType
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import Tracklet


def distance(name: str, x: np.ndarray, y: np.ndarray) -> float:
    """
    Calculates distance between two vectors. Supports: L1 and L2.

    Args:
        name: Distance name
        x: Vector X
        y: Vector Y

    Returns:
        Distance between x and y.
    """
    if name == 'l1':
        return np.abs(x - y).sum()
    if name == 'l2':
        return np.sqrt(np.square(x - y).sum())

    raise AssertionError('Invalid Program State!')


@ASSOCIATION_CATALOG.register('distance')
class DistanceAssociation(AssociationAlgorithm):
    """
    Calculates distance between detection and prediction relative to the last track position.
    """
    def __init__(
        self,
        distance_name: str = 'l1',
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            distance_name: Distance function mae
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)
        self._distance_name = distance_name


    def form_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        _ = object_features  # Unused

        n_tracklets, n_detections = len(tracklet_estimations), len(detections)
        cost_matrix = np.zeros(shape=(n_tracklets, n_detections), dtype=np.float32)
        for t_i in range(n_tracklets):
            tracklet_estimated_bbox = tracklet_estimations[t_i]
            tracklet_info = tracklets[t_i]

            tracklet_last_bbox = tracklet_info.bbox.as_numpy_xyxy()
            tracklet_motion = tracklet_estimated_bbox.as_numpy_xyxy() - tracklet_last_bbox

            for d_i in range(n_detections):
                det_bbox = detections[d_i]
                det_motion = det_bbox.as_numpy_xyxy() - tracklet_last_bbox
                cost_matrix[t_i][d_i] = distance(self._distance_name, tracklet_motion, det_motion)

        return cost_matrix


@ASSOCIATION_CATALOG.register('move')
class Move(ComposeAssociationAlgorithm):
    """
    Combines Hungarian IOU and motion estimation for tracklet and detection matching.
    """
    def __init__(
        self,
        match_threshold: float = 0.30,
        motion_lambda: float = 5,
        distance_name: str = 'l1',
        label_gating: Optional[LabelGatingType] = None,
        fuse_score: bool = False,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            match_threshold: IOU match gating
            motion_lambda: Motion difference multiplier
            label_gating: Gating between different types of objects
            fuse_score: Fuse Hungarian IoU score
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        matchers = [
            IoUAssociation(
                match_threshold=match_threshold,
                label_gating=label_gating,
                fuse_score=fuse_score
            ),
            DistanceAssociation(
                distance_name=distance_name
            )
        ]
        weights = [1, motion_lambda]

        super().__init__(
            matchers=matchers,
            weights=weights,
            fast_linear_assignment=fast_linear_assignment
        )



"""
Standard SORT association method based solely on IoU.
"""
from typing import Optional, Union, List, Tuple

import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import Tracklet, TrackletState

LabelType = Union[int, str]
LabelGatingType = Union[LabelType, List[Tuple[LabelType, LabelType]]]


class IoUBasedAssociation(AssociationAlgorithm):
    """
    Standard negative IoU association with gating.
    """

    def __init__(
        self,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            label_gating: Define which object labels can be matched
                - If not defined, matching is label agnostic
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)

        self._label_gating = {tuple(e) for e in label_gating} if label_gating is not None else None

    def _can_match_labels(self, tracklet_label: LabelType, det_label: LabelType) -> bool:
        """
        Checks if matching between tracklet and detection is possible.

        Args:
            tracklet_label: Tracklet label
            det_label: Detection label

        Returns:
            True if matching is possible else False
        """
        if tracklet_label == det_label:
            # Objects with same label can always match
            return True

        if self._label_gating is None:
            # If label gating is not set then any objects with same label can't match
            return False

        return (tracklet_label, det_label) in self._label_gating \
            or (det_label, tracklet_label) in self._label_gating

    def form_cost_matrix(
            self,
            tracklet_estimations: List[PredBBox],
            detections: List[PredBBox],
            object_features: Optional[np.ndarray] = None,
            tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        _ = object_features  # Unused

        n_tracklets, n_detections = len(tracklet_estimations), len(detections)
        cost_matrix = np.zeros(shape=(n_tracklets, n_detections), dtype=np.float32)
        for t_i in range(n_tracklets):
            tracklet = tracklets[t_i]
            tracklet_bbox = tracklet_estimations[t_i]

            for d_i in range(n_detections):
                det_bbox = detections[d_i]

                # Check if matching is possible
                if not self._can_match_labels(tracklet_bbox.label, det_bbox.label):
                    cost_matrix[t_i][d_i] = np.inf
                    continue

                cost_matrix[t_i][d_i] = self._calc_score(tracklet, tracklet_bbox, det_bbox)

        return cost_matrix

    def _calc_score(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox) -> float:
        """
        Calculates matching score (cost) between tracklet and a detection bounding box.

        Args:
            tracklet: Tracklet info
            tracklet_bbox: Tracklet bounding box
            det_bbox: Detection bounding box

        Returns:
            Matching score
        """
        raise NotImplementedError('Score calculation is not implemented!')


@ASSOCIATION_CATALOG.register('iou')
class IoUAssociation(IoUBasedAssociation):
    """
    Standard negative IoU association with gating.
    """
    def __init__(
        self,
        match_threshold: float = 0.30,
        fuse_score: bool = False,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            match_threshold: Min threshold do match tracklet with object.
                If threshold is not met then cost is equal to infinity.
            fuse_score: Fuse score with iou
            label_gating: Define which object labels can be matched
                - If not defined, matching is label agnostic
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(
            label_gating=label_gating,
            fast_linear_assignment=fast_linear_assignment
        )

        self._match_threshold = match_threshold
        self._fuse_score = fuse_score

    def _calc_score(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox) -> float:
        _ = tracklet

        # Calculate IOU score
        iou_score = tracklet_bbox.iou(det_bbox)
        # Higher the IOU score the better is the match (using negative values because of min optim function)
        # If score has very high value then
        score = iou_score * det_bbox.conf if self._fuse_score else iou_score
        return - score if score > self._match_threshold else np.inf


@ASSOCIATION_CATALOG.register('adaptive-iou')
class IoUAssociationAdaptiveGating(IoUBasedAssociation):
    """
    Adaptive IoU scales the match threshold based on the object's velocity.
    Object's that have less movement should have more overlap in the next frame.
    """
    def __init__(
        self,
        match_threshold: float = 0.30,
        fuse_score: bool = False,
        gating_factor: float = 1.0,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            match_threshold: Min threshold do match tracklet with object.
                If threshold is not met then cost is equal to infinity.
            fuse_score: Fuse score with iou
            gating_factor: Adaptive gating factor
            label_gating: Define which object labels can be matched
                - If not defined, matching is label agnostic
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(
            label_gating=label_gating,
            fast_linear_assignment=fast_linear_assignment
        )

        self._match_threshold = match_threshold
        self._fuse_score = fuse_score
        self._gating_factor = gating_factor

    def _calc_score(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox) -> float:
        # Estimate variable gating threshold
        match_threshold = self._match_threshold
        if len(tracklet.history) >= 2:
            last_bbox_index, last_bbox = tracklet.history[-1]
            prev_bbox_index, prev_bbox = tracklet.history[-2]
            if last_bbox_index > prev_bbox_index:
                movement = np.abs(last_bbox.as_numpy_xyxy() - prev_bbox.as_numpy_xyxy() / (last_bbox_index - prev_bbox_index)).mean()
                match_threshold = max(0, match_threshold - movement * self._gating_factor)

        iou_score = tracklet_bbox.iou(det_bbox)
        score = iou_score * det_bbox.conf if self._fuse_score else iou_score
        return - score if score > match_threshold else np.inf


@ASSOCIATION_CATALOG.register('hmiou')
class HMIoUAssociation(IoUBasedAssociation):
    """
    Height modulated IoU. Reference: https://arxiv.org/pdf/2308.00783.pdf
    """
    def __init__(
        self,
        match_threshold: float = 0.30,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            match_threshold: Min threshold do match tracklet with object.
                If threshold is not met then cost is equal to infinity.
            label_gating: Define which object labels can be matched
                - If not defined, matching is label agnostic
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(
            label_gating=label_gating,
            fast_linear_assignment=fast_linear_assignment
        )

        self._match_threshold = match_threshold

    @staticmethod
    def hiou(tracklet_bbox: PredBBox, det_bbox: PredBBox) -> float:
        """
        Calculates HIoU between two bboxes.

        Args:
            tracklet_bbox: Tracklet bounding box
            det_bbox: Detection bounding box

        Returns:
            HIoU between two bboxes
        """
        intersection_bottom = min(tracklet_bbox.bottom_right.y, det_bbox.bottom_right.y)
        intersection_top = max(tracklet_bbox.upper_left.y, det_bbox.upper_left.y)
        union_bottom = max(tracklet_bbox.bottom_right.y, det_bbox.bottom_right.y)
        union_top = min(tracklet_bbox.upper_left.y, det_bbox.upper_left.y)

        if union_top >= union_bottom or intersection_top >= intersection_bottom:
            return 0.0

        return (intersection_bottom - intersection_top) / (union_bottom - union_top)

    def _calc_score(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox) -> float:
        _ = tracklet

        # Calculate IOU score
        iou_score = tracklet_bbox.iou(det_bbox)
        height_iou_score = self.hiou(tracklet_bbox, det_bbox)
        return - iou_score * height_iou_score * height_iou_score if iou_score > self._match_threshold else np.inf


@ASSOCIATION_CATALOG.register('decay-iou')
class DecayIoU(IoUBasedAssociation):
    """
    IoU but with threshold decay.
    """
    def __init__(
        self,
        min_threshold: float = 0.30,
        max_threshold: float = 0.50,
        threshold_decay: float = 0.02,
        fuse_score: bool = False,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            min_threshold: Min IoU threshold
            max_threshold: Initial IoU threshold
            threshold_decay: IoU threshold decay
            fuse_score: Fuse score with iou
            label_gating: Define which object labels can be matched
                - If not defined, matching is label agnostic
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(
            label_gating=label_gating,
            fast_linear_assignment=fast_linear_assignment
        )

        self._min_threshold = min_threshold
        self._max_threshold = max_threshold
        self._threshold_decay = threshold_decay
        self._fuse_score = fuse_score

    def _calc_score(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox) -> float:
        _ = tracklet

        # Calculate IOU score
        iou_score = tracklet_bbox.iou(det_bbox)
        # Higher the IOU score the better is the match (using negative values because of min optim function)
        # If score has very high value then
        score = iou_score * det_bbox.conf if self._fuse_score else iou_score

        match_threshold = max(self._min_threshold, self._max_threshold - self._threshold_decay * tracklet.lost_time)
        return - score if score > match_threshold else np.inf




"""
Tracker inference visualization.
"""
import logging
import os

import hydra
from omegaconf import DictConfig

from motrack.common.project import CONFIGS_PATH
from motrack.config_parser import GlobalConfig
from motrack.datasets import dataset_factory
from motrack.utils import pipeline
from motrack.tools import run_visualize_tracker_inference

logger = logging.getLogger('TrackerVizualization')


@pipeline.task('visualize-inference')
def visualize_inference(cfg: GlobalConfig) -> None:
    tracker_output_option = os.path.join(cfg.experiment_path, cfg.visualize.option)
    tracker_output_active = os.path.join(cfg.experiment_path, 'active')
    assert os.path.exists(cfg.experiment_path), f'Path "{cfg.experiment_path}" does not exist!'
    logger.info(f'Visualizing tracker inference on path "{tracker_output_option}".')

    dataset = dataset_factory(
        dataset_type=cfg.dataset.type,
        path=cfg.dataset.fullpath,
        params=cfg.dataset.params,
        test=cfg.eval.split == 'test'
    )

    run_visualize_tracker_inference(
        dataset=dataset,
        tracker_active_output=tracker_output_active,
        tracker_output_option=tracker_output_option,
        visualize_cfg=cfg.visualize,
        scene_pattern=cfg.dataset_filter.scene_pattern
    )

@hydra.main(config_path=CONFIGS_PATH, config_name='movesort', version_base='1.1')
def main(cfg: DictConfig):
    # noinspection PyTypeChecker
    visualize_inference(cfg)


if __name__ == '__main__':
    main()



"""
Implementation of SparseTrack DCM association method.
"""
from typing import Optional, List, Tuple

import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.algorithms.iou import IoUAssociation, LabelGatingType
from motrack.tracker.matching.algorithms.move import Move
from motrack.tracker.tracklet import Tracklet
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG


class DCM(AssociationAlgorithm):
    """
    SparseTrack DCM: https://arxiv.org/pdf/2306.05238.pdf
    """
    def __init__(
        self,
        matcher: AssociationAlgorithm,
        levels: int = 12,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            matcher: Base matcher
            levels: Number of depth levels to use
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)

        self._matcher = matcher
        self._levels = levels

    def _create_depth_masks(self, bboxes: List[PredBBox]) -> np.ndarray:
        """
        Arranges list of bboxes by levels.

        Args:
            bboxes: List of tracklet or detection bboxes

        Returns:
            Depth level class for each bbox
        """
        n_bboxes = len(bboxes)
        masks = np.full(shape=(self._levels, n_bboxes), fill_value=True, dtype=bool)
        if n_bboxes == 0:
            return masks

        neg_bottoms = [-bbox.bottom_right.y for bbox in bboxes]  # Pedestrian bbox bottom line
        min_bottom, max_bottom = min(neg_bottoms), max(neg_bottoms)
        depths_range = np.linspace(min_bottom, max_bottom, self._levels + 1, endpoint=True)

        prev_depth = depths_range[0]
        for i, depth in enumerate(depths_range[1:]):
            depth: float

            for j in range(n_bboxes):
                if i + 2 < depths_range.shape[0]:
                    masks[i, j] = prev_depth <= neg_bottoms[j] < depth
                else:
                    masks[i, j] = prev_depth <= neg_bottoms[j] <= depth

            prev_depth = depth

        assert masks.any(axis=0).all(), f'Invalid Mask! Info:\n{masks=}\n{masks.any(axis=1)=}\n{depths_range=}\n{neg_bottoms=}\n{bboxes}'
        return masks

    def match(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:
        _ = object_features  # Unused

        t_masks = self._create_depth_masks(tracklet_estimations)
        d_masks = self._create_depth_masks(detections)

        unmatched_tracklet_indices: List[int] = []
        unmatched_detection_indices: List[int] = []
        matches: List[Tuple[int, int]] = []
        for tm, dm in zip(t_masks, d_masks):
            # Combine unmatched tracklets and detections with current DCM mask level
            level_tracklet_indices = [t_i for t_i, include in enumerate(tm) if include] + unmatched_tracklet_indices
            level_detection_indices = [d_i for d_i, include in enumerate(dm) if include] + unmatched_detection_indices

            # Extract info
            level_tracklet_estimations = [tracklet_estimations[t_i] for t_i in level_tracklet_indices]
            level_tracklets = [tracklets[t_i] for t_i in level_tracklet_indices] if tracklets is not None else None
            level_detections = [detections[d_i] for d_i in level_detection_indices]

            # Perform matching
            level_matches, unmatched_tracklet_indices, unmatched_detection_indices = self._matcher(
                level_tracklet_estimations, level_detections, tracklets=level_tracklets
            )

            # Map "level" indices to real indices
            level_matches = [(level_tracklet_indices[t_i], level_detection_indices[d_i]) for t_i, d_i in level_matches]
            unmatched_tracklet_indices = [level_tracklet_indices[t_i] for t_i in unmatched_tracklet_indices]
            unmatched_detection_indices = [level_detection_indices[d_i] for d_i in unmatched_detection_indices]

            matches.extend(level_matches)

        return matches, unmatched_tracklet_indices, unmatched_detection_indices


@ASSOCIATION_CATALOG.register('dcm')
class DCMIoU(DCM):
    """
    SparseTrack DCM (original): https://arxiv.org/pdf/2306.05238.pdf
    """
    def __init__(
        self,
        levels: int = 12,
        match_threshold: float = 0.30,
        fuse_score: bool = False,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            match_threshold: IoU gating match threshold
            fuse_score: Multiply IoU score by detection score
            levels: Number of depth levels to use
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        matcher = IoUAssociation(
            match_threshold=match_threshold,
            label_gating=label_gating,
            fuse_score=fuse_score
        )
        super().__init__(
            matcher=matcher,
            levels=levels,
            fast_linear_assignment=fast_linear_assignment
        )


@ASSOCIATION_CATALOG.register('move-dcm')
class MoveDCM(DCM):
    """
    DCM + Move
    """
    def __init__(
        self,
        levels: int = 12,
        match_threshold: float = 0.30,
        motion_lambda: float = 5,
        distance_name: str = 'l1',
        label_gating: Optional[LabelGatingType] = None,
        fuse_score: bool = False,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            levels: Number of depth levels to use
            match_threshold: IoU gating match threshold
            motion_lambda: Move cost multiplier
            fuse_score: Multiply IoU score by detection score
            levels: Number of depth levels to use
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        matcher = Move(
            match_threshold=match_threshold,
            motion_lambda=motion_lambda,
            distance_name=distance_name,
            label_gating=label_gating,
            fuse_score=fuse_score,
            fast_linear_assignment=fast_linear_assignment
        )
        super().__init__(
            matcher=matcher,
            levels=levels,
        )



"""
RAII wrapper of MP4Writer.
"""
from pathlib import Path
from typing import Tuple, Optional

import cv2
import numpy as np


class MP4WriterResolutionMismatchException(Exception):
    """
    Frame resolution does not match with the previously used one
    """
    pass


class MP4WriterNotOpen(Exception):
    """
    MP4Writer is not open yet
    """
    pass


class MP4Writer:
    """
    RAII MP4 writer
    """
    # noinspection PyUnresolvedReferences
    def __init__(
        self,
        path: str,
        fps: int,
        shape: Optional[Tuple[int, int]] = None,
        resize: bool = False,
        resize_interpolation: int = cv2.INTER_NEAREST
    ):
        """
        Args:
            path: Output video path
            fps: Output video fps
            shape: Set video output shape (optional) - Format (w, h)
                - if not set then first video frame resolution will be taken as shape
            resize: Resize video to some shape (requires shape to be set)
        """
        self._fps = fps
        self._path = path
        self._resize = resize
        self._resize_interpolation = resize_interpolation

        # State
        self._writer = None  # Stored on first frame
        self._shape: Optional[Tuple[int, int]] = shape  # Stored on first frame if not set

        # Check
        if self._resize and not self._shape:
            raise AttributeError('Shape has to be set with the "resize" option!')

    @property
    def fps(self) -> int:
        """
        Gets video fps.

        Returns:
            fps
        """
        return self._fps

    @property
    def resolution(self) -> Tuple[int, int]:
        """
        Returns:
            Video resolution
        """
        if self._shape is None:
            raise MP4WriterNotOpen('Shape is unknown. Please use write() at least once.')
        return self._shape

    def write(self, frame: np.ndarray) -> None:
        """
        Appends new frame to mp4 video.

        Args:
            frame: Frame to write to video
        """
        h, w, _ = frame.shape
        shape = (w, h)
        if self._writer is None:
            if self._shape is None:
                self._shape = shape

            # noinspection PyUnresolvedReferences
            fourcc = cv2.VideoWriter_fourcc(*'mp4v')
            Path(self._path).parent.mkdir(parents=True, exist_ok=True)
            # noinspection PyUnresolvedReferences
            self._writer = cv2.VideoWriter(self._path, fourcc, self._fps, self._shape)

        if shape != self._shape:
            if not self._resize:
                raise MP4WriterResolutionMismatchException(f'Resolution mismatch! Expected {self._shape} but got {shape}.')
            else:
                # noinspection PyUnresolvedReferences
                frame = cv2.resize(frame, self._shape, interpolation=self._resize_interpolation)

        self._writer.write(frame)

    def close(self) -> None:
        """
        Closes mp4 writer.
        """
        self._writer.release()

    def __enter__(self) -> 'MP4Writer':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Association method interface.
"""
from typing import List, Optional, Tuple

import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.matching.utils import hungarian, greedy
from motrack.tracker.tracklet import Tracklet


class AssociationAlgorithm:
    """
    Defines interface to tracklets-detections association matching.
    """
    def __init__(
        self,
        fast_linear_assignment: bool = False,
    ):
        """
        fast_linear_assignment: Use greedy algorithm for linear assignment
            - This might be more efficient in case of large cost matrix
        """
        self._fast_linear_assignment = fast_linear_assignment

    def form_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        """
        Forms matching cost matrix used as the input to the linear assignment solver.

        Args:
            tracklet_estimations: Tracked object from previous frames
            detections: Currently detected objects
            object_features: Object appearance features (optional)
            tracklets: Full tracklet info (optional)

        Returns:
            Cost matrix
        """
        raise NotImplementedError('Form cost matrix is not implemented!')

    def match(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:
        """
        Performs matching between tracklets and detections (interface).

        Args:
            tracklet_estimations: Tracked object from previous frames
            detections: Currently detected objects
            object_features: Object appearance features (optional)
            tracklets: Full tracklet info (optional)

        Returns:
            - List of matches (pairs)
            - list of unmatched tracklets
            - list of unmatched detections
        """
        cost_matrix = self.form_cost_matrix(
            tracklet_estimations=tracklet_estimations,
            detections=detections,
            object_features=object_features,
            tracklets=tracklets
        )

        return greedy(cost_matrix) if self._fast_linear_assignment else \
            hungarian(cost_matrix)

    def __call__(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:
        return self.match(tracklet_estimations, detections, object_features=object_features, tracklets=tracklets)



"""
Implementation of Sort tracker with a custom filter.
"""
from typing import Optional, Dict, Any, List

import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.matching import association_factory
from motrack.tracker.trackers.algorithms.motion_reid import MotionReIDBasedTracker
from motrack.tracker.trackers.catalog import TRACKER_CATALOG
from motrack.tracker.tracklet import Tracklet


@TRACKER_CATALOG.register('sort')
class SortTracker(MotionReIDBasedTracker):
    """
    Baseline Sort tracker components:
    - ReID: HungarianIOU
    - Motion model: Filter prior
    - Combining detection and motion model: Filter posterior
    """
    def __init__(
        self,
        filter_name: str = 'bot-sort',
        filter_params: Optional[dict] = None,
        cmc_name: Optional[str] = None,
        cmc_params: Optional[dict] = None,
        reid_name: Optional[str] = None,
        reid_params: Optional[str] = None,
        matcher_algorithm: str = 'iou',
        matcher_params: Optional[Dict[str, Any]] = None,
        remember_threshold: int = 1,
        initialization_threshold: int = 3,
        new_tracklet_detection_threshold: Optional[float] = None,
        use_observation_if_lost: bool = False,
        appearance_ema_momentum: float = 0.95,
        appearance_buffer: int = 0
    ):
        """
        Args:
            filter_name: Filter name
            filter_params: Filter params
            cmc_name: CMC name
            cmc_params: CMC params
            reid_name: ReID name
            reid_params: ReID params

            matcher_algorithm: Choose matching algorithm (e.g. Hungarian IOU)
            matcher_params: Matching algorithm parameters

            remember_threshold: How long does the tracklet without any detection matching
                If tracklet isn't matched for more than `remember_threshold` frames then
                it is deleted.
            initialization_threshold: Number of frames until tracklet becomes active
            new_tracklet_detection_threshold: Threshold to accept new tracklet
            use_observation_if_lost: When re-finding tracklet, use observation instead of estimation
            appearance_buffer: Appearance buffer length (set to a non-zero non-negative integer to activate)
        """
        if filter_params is None:
            filter_params = {}

        super().__init__(
            filter_name=filter_name,
            filter_params=filter_params,
            cmc_name=cmc_name,
            cmc_params=cmc_params,
            reid_name=reid_name,
            reid_params=reid_params,
            reid_detection_threshold=None,

            new_tracklet_detection_threshold=new_tracklet_detection_threshold,
            remember_threshold=remember_threshold,
            initialization_threshold=initialization_threshold,
            use_observation_if_lost=use_observation_if_lost,

            appearance_ema_momentum=appearance_ema_momentum,
            appearance_buffer=appearance_buffer
        )

        matcher_params = {} if matcher_params is None else matcher_params
        self._matcher = association_factory(name=matcher_algorithm, params=matcher_params)

    def _track(
        self,
        tracklets: List[Tracklet],
        prior_tracklet_bboxes: List[PredBBox],
        detections: List[PredBBox],
        frame_index: int,
        objects_features: Optional[np.ndarray] = None,
        frame: Optional[np.ndarray] = None
    ) -> List[Tracklet]:
        # Perform matching
        matches, unmatched_tracklets, unmatched_detections = self._matcher(prior_tracklet_bboxes, detections,
                                                                           object_features=objects_features, tracklets=tracklets)

        self._update_tracklets(
            matched_tracklets=[tracklets[t_i] for t_i, _ in matches],
            matched_detections=[detections[d_i] for _, d_i in matches],
            matched_object_features=objects_features[[d_i for _, d_i in matches]] if objects_features is not None else None,
            frame_index=frame_index
        )

        # Create new tracklets from unmatched detections and initiate filter states
        new_tracklets = self._create_new_tracklets(
            detections=[detections[d_i] for d_i in unmatched_detections],
            frame_index=frame_index,
            objects_features=objects_features[unmatched_detections] if objects_features is not None else None
        )

        # Delete old and new unmatched tracks
        self._handle_lost_tracklets(
            lost_tracklets=[tracklets[t_i] for t_i in unmatched_tracklets]
        )

        return tracklets + new_tracklets



"""
ReID factory method.
Use `REID_CATALOG.register` to extend supported ReID algorithms.
"""
from motrack.reid.algorithms.base import BaseReID
# noinspection PyUnresolvedReferences
from motrack.reid.algorithms.fastreid_onnx import FastReIDOnnx  # pylint: disable=unused-import
from motrack.reid.catalog import REID_CATALOG


def reid_inference_factory(
    name: str,
    params: dict,
) -> BaseReID:
    """
    Creates ReID inference.

    Args:
        name: ReID inference name (type)
        params: CTor parameters

    Returns:
        Initialized ReID inference
    """
    return REID_CATALOG[name](**params)



"""
Implementation of C-BIoU and BIoU association methods.
"""
from typing import Optional, List, Tuple

import numpy as np

from motrack.library.cv.bbox import PredBBox, BBox
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.algorithms.iou import IoUAssociation, LabelGatingType
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import Tracklet


@ASSOCIATION_CATALOG.register('biou')
class HungarianBIoU(IoUAssociation):
    """
    BIoU matching algorithm. Ref: https://arxiv.org/pdf/2211.14317.pdf
    """
    def __init__(
        self,
        b: float = 0.3,
        match_threshold: float = 0.30,
        label_gating: Optional[LabelGatingType] = None,
        *args, **kwargs
    ):
        """
        Args:
            b: BBox buffer
            match_threshold: IOU match gating
            label_gating: Gating between different types of objects
        """
        super().__init__(
            match_threshold=match_threshold,
            label_gating=label_gating,
            *args, **kwargs
        )
        self._b = b

    def _buffer_bbox(self, bbox: PredBBox) -> PredBBox:
        center = bbox.center
        w, h = bbox.width, bbox.height
        w *= (1 + 2 * self._b)
        h *= (1 + 2 * self._b)

        return PredBBox.create(
            bbox=BBox.from_cxywh(center.x, center.y, w, h),
            label=bbox.label,
            conf=bbox.conf
        )

    def form_iou_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        _, _ = object_features, tracklets  # Unused

        tracklet_estimations = [self._buffer_bbox(bbox) for bbox in tracklet_estimations]
        detections = [self._buffer_bbox(bbox) for bbox in detections]
        return self.form_iou_cost_matrix(tracklet_estimations, detections)


@ASSOCIATION_CATALOG.register('cbiou')
class HungarianCBIoU(AssociationAlgorithm):
    """
    C-BIoU matching algorithm. Ref: https://arxiv.org/pdf/2211.14317.pdf
    """
    def __init__(
        self,
        b1: float = 0.3,
        b2: float = 0.4,
        match_threshold: float = 0.30,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False,
    ):
        """
        Args:
            b1: First buffer matching threshold
            b2: Second buffer matching threshold
            match_threshold: IoU match gating
            label_gating: Gating between different types of objects
            fast_linear_assignment: Use greedy algorithm for linear assignment
            - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)

        self._biou1_matcher = HungarianBIoU(
            b=b1,
            match_threshold=match_threshold,
            label_gating=label_gating,
        )

        self._biou1_matcher = HungarianBIoU(
            b=b2,
            match_threshold=match_threshold,
            label_gating=label_gating,
        )

    def match(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:
        _ = object_features  # Unused

        # First matching
        matches1, unmatched_tracklet_indices1, unmatched_detection_indices1 = \
            self._biou1_matcher(tracklet_estimations, detections, tracklets=tracklets)
        unmatched_tracklet_estimations = [tracklet_estimations[t_i] for t_i in unmatched_tracklet_indices1]
        unmatched_tracklets = [tracklets[t_i] for t_i in unmatched_tracklet_indices1]
        unmatched_detections = [detections[d_i] for d_i in unmatched_detection_indices1]

        # Second matching
        matches2, unmatched_tracklet_indices2, unmatched_detection_indices2 = \
            self._biou1_matcher(unmatched_tracklet_estimations, unmatched_detections, tracklets=unmatched_tracklets)
        unmatched_tracklet_indices = [unmatched_tracklet_indices1[t_i] for t_i in unmatched_tracklet_indices2]
        unmatched_detection_indices = [unmatched_detection_indices1[t_i] for t_i in unmatched_detection_indices2]
        matches2 = [(unmatched_tracklet_indices1[t_i], unmatched_detection_indices1[d_i]) for t_i, d_i in matches2]
        matches = matches1 + matches2

        return matches, unmatched_tracklet_indices, unmatched_detection_indices



"""
CMC factory method.
Use `CMC_CATALOG.register` to extend supported CMC algorithms.
"""
from motrack.cmc.algorithms.base import CameraMotionCompensation
# noinspection PyUnresolvedReferences
from motrack.cmc.algorithms.gmc_from_file import GMCFromFile  # pylint: disable=unused-import
from motrack.cmc.catalog import CMC_CATALOG


def cmc_factory(name: str, params: dict) -> CameraMotionCompensation:
    """
    CMC factory.

    Args:
        name: Algorithm name
        params: Dataset creation parameters

    Returns:
        Created CMC object
    """
    return CMC_CATALOG[name](**params)



"""
Tracker association factory method.
Use `ASSOCIATION_CATALOG.register` to extend supported tracker association algorithms.
"""
# pylint: disable=unused-import
import copy
from typing import Dict, Any, List

from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.biou import HungarianCBIoU, HungarianBIoU
from motrack.tracker.matching.algorithms.compose import ComposeAssociationAlgorithm
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.conf import HybridConfidenceAssociation
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.dcm import DCMIoU, MoveDCM
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.horizontal_cues import HorizontalViewCues
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.iou import IoUAssociation, IoUAssociationAdaptiveGating, HMIoUAssociation, DecayIoU
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.momentum import OCM, RobustOCM, ECM, RobustECM
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.move import Move
# noinspection PyUnresolvedReferences
from motrack.tracker.matching.algorithms.reid import ReIDIoUAssociation
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG


def association_factory(name: str, params: Dict[str, Any]) -> AssociationAlgorithm:
    """
    Association algorithm factory. Not case-sensitive.

    Allows weighted composition of multiple association algorithms
    (that implement the `_form_cost_matrix` method). Format:
    ```
    algorithm_name: compose
    algorithm_params:
        matchers:
            - name: iou
              params:
                match_threshold: 0.3
            - name: move
              params:
                match_threshold: 0.2
                motion_lambda: 0.1
        weights: [0.3, 0.7]
    ```

    Args:
        name: Algorithm name
        params: Parameters

    Returns:
        AssociationAlgorithm object
    """
    if name == 'compose':
        params = copy.deepcopy(params)

        assert 'matchers' in params, 'Expected "matchers" to be defined in parameters'
        matchers = params.pop('matchers')
        assert isinstance(matchers, list), f'Expected list for "matchers" by found {type(matchers)} instead!'

        assert 'weights' in params, 'Expected "weights" to be defined in parameters'
        weights = params.pop('weights')

        matcher_objs: List[AssociationAlgorithm] = []
        for matcher in matchers:
            combined_params = {**matcher['params'], **params}
            matcher_obj = association_factory(matcher['name'], combined_params)
            matcher_objs.append(matcher_obj)

        return ComposeAssociationAlgorithm(
            matchers=matcher_objs,
            weights=weights
        )

    return ASSOCIATION_CATALOG[name](**params)



"""
DetectionManager makes tracker development easier and faster
by adding cache and oracle detections options.
"""
import os
from typing import Optional, List, Tuple, Union

import numpy as np

from motrack.datasets import BaseDataset
from motrack.library.cv.bbox import PredBBox
from motrack.library.numpy_utils.io import load_npy, store_npy
from motrack.object_detection.factory import object_detection_inference_factory
from motrack.utils.lookup import LookupTable


class DetectionManager:
    """
    Detection manager integrates multiple features, making the usage flexible. Features:
    - Performs inference
    - Uses cache for faster multiple runs
    - Uses optional `oracle` GT detections used for debugging and testing
    """
    def __init__(
        self,
        inference_name: str,
        inference_params: dict,
        dataset: BaseDataset,
        lookup: Optional[LookupTable] = None,
        cache_path: Optional[str] = None,
        oracle: bool = False
    ):
        """
        Args:
            inference_name: Inference model name
            inference_params: Inference model parameters
            dataset: Dataset
            lookup: Lookup (if set then returns string labels instead of ints)
            cache_path: If set then cache is used
            oracle: Use GT predictions instead of running inference
        """
        self._inference = object_detection_inference_factory(name=inference_name, params=inference_params, lookup=lookup)
        self._dataset = dataset
        self._lookup = lookup
        self._cache_path = cache_path
        self._oracle = oracle

    def predict(self, scene: str, frame_index: int) -> List[PredBBox]:
        """
        Performs prediction for given scene name and frame index.

        If `cache_path` is set then inference is faster after first run.
        If `oracle` is set to True then ground truth is used instead of inference

        Args:
            scene: Scene name
            frame_index: Frame inde

        Returns:
            List of inference bboxes
        """
        if self._oracle:
            bboxes, classes, confidences = self._oracle_inference(scene, frame_index)
        else:
            scene_frame_cache_path, _, _, _ = self._get_cache_paths(scene, frame_index)
            if self._cache_path is not None and os.path.exists(scene_frame_cache_path):
                bboxes, classes, confidences = self._load_from_cache(scene, frame_index)
            else:
                image = self._dataset.load_scene_image_by_frame_index(scene, frame_index)
                bboxes, classes, confidences = self._inference.predict_with_postprocess(image)
                if self._cache_path is not None:
                    self._store_cache(scene, frame_index, bboxes, classes, confidences)

        return self._inference.pack_bboxes(bboxes, classes, confidences)

    def _load_from_cache(self, scene: str, frame_index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Loads inference info (bboxes, classes, confidences) from cache.

        Args:
            scene: Scene name
            frame_index: Frame index

        Returns:
            Cached inference data
        """
        _, bboxes_cache_path, classes_cache_path, confidences_cache_path = self._get_cache_paths(scene, frame_index)
        return load_npy(bboxes_cache_path), load_npy(classes_cache_path), load_npy(confidences_cache_path)

    def _store_cache(self, scene: str, frame_index: int, bboxes: np.ndarray, classes: np.ndarray, confidences: np.ndarray) -> None:
        """
        Stores cache for faster inference in next iteration

        Args:
            scene: Scene name
            frame_index: Frame index
            bboxes: BBoxes data (numpy)
            classes: Classes data (numpy)
            confidences: Confidences data (numpy)
        """
        _, bboxes_cache_path, classes_cache_path, confidences_cache_path = self._get_cache_paths(scene, frame_index)
        store_npy(bboxes, bboxes_cache_path)
        store_npy(classes, classes_cache_path)
        store_npy(confidences, confidences_cache_path)

    def _get_cache_paths(self, scene: str, frame_index: int) -> Tuple[str, str, str, str]:
        """
        Get path where the inference cache can be found

        Args:
            scene: Scene name
            frame_index: Frame index

        Returns:
            Scene frame cache path
        """
        scene_frame_cache_path = os.path.join(self._cache_path, scene, f'{frame_index:06d}')
        bboxes_cache_path = os.path.join(scene_frame_cache_path, 'bboxes.npy')
        classes_cache_path = os.path.join(scene_frame_cache_path, 'classes.npy')
        confidences_cache_path = os.path.join(scene_frame_cache_path, 'confidences.npy')
        return scene_frame_cache_path, bboxes_cache_path, classes_cache_path, confidences_cache_path

    def _oracle_inference(self, scene: str, frame_index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns GT prediction (annotations)

        Args:
            scene: Scene name
            frame_index: Frame index

        Returns:
            Oracle detections
        """
        object_ids = self._dataset.get_scene_object_ids(scene)
        frame_bboxes: List[np.ndarray] = []
        frame_classes: List[Union[int, float]] = []
        frame_confidences: List[float] = []

        for object_id in object_ids:
            data = self._dataset.get_object_data_by_frame_index(object_id, frame_index, relative_bbox_coords=True)
            if data is None:
                continue

            bbox = data.bbox
            if isinstance(bbox, list):
                bbox = np.array(bbox, dtype=np.float32)
            bbox[[2, 3]] += bbox[[0, 1]]
            cls = self._lookup[data.category]

            frame_bboxes.append(bbox)
            frame_classes.append(cls)
            frame_confidences.append(1.0)

        bboxes = np.stack(frame_bboxes)
        classes = np.array(frame_classes, dtype=np.float32)
        confidences = np.array(frame_confidences, dtype=np.float32)
        return bboxes, classes, confidences



"""
Wrapper for Motrack-motion end-to-end filter.
"""
from typing import Tuple

import numpy as np
import torch
from motrack.filter.algorithms.base import StateModelFilter, State
from motrack.filter.catalog import FILTER_CATALOG
from motrack_motion.datasets import transforms
from motrack_motion.filter.factory import filter_factory
from motrack_motion.models import model_factory
from torch import nn


@FILTER_CATALOG.register('motrack-motion-end-to-end')
class MotrackMotionFilterEndToEnd(StateModelFilter):
    """
    Motrack-motion framework end-to-end filters
    """
    def __init__(
        self,
        name: str,
        params: dict,
        model_type: str,
        model_params: dict,
        checkpoint_path: str,
        transform_type: str,
        transform_params: dict
    ):
        """
        Args:
            name: Filter name
            params: Filter params
            model_type: Model type (architecture name)
            model_params: Model parameters
            checkpoint_path: Checkpoint path (PLT checkpoint)
            transform_type: Transform type
            transform_params: Transform parameters
        """
        transform = transforms.transform_factory(transform_type, transform_params)

        self._core = filter_factory(
            name=name,
            params=params,
            model=self._load_model_from_pl_checkpoint(model_type, model_params, checkpoint_path),
            transform=transform,
        )

    @staticmethod
    def _load_model_from_pl_checkpoint(
            model_type: str,
        model_params: dict,
        checkpoint_path: str
    ) -> nn.Module:
        """
        Loads model weights from a PLT checkpoint.

        Args:
            model_type: Model type (architecture name)
            model_params: Model parameters
            checkpoint_path: Checkpoint path (PLT checkpoint)

        Returns:
            Loaded model
        """
        model = model_factory(model_type, model_params)
        state_dict = torch.load(checkpoint_path)
        model_state_dict = {k.replace('_model.', '', 1): v for k, v in state_dict['state_dict'].items()}
        model.load_state_dict(model_state_dict)
        return model

    def initiate(self, measurement: np.ndarray) -> State:
        measurement = torch.from_numpy(measurement)
        return self._core.initiate(measurement)

    def predict(self, state: State) -> State:
        return self._core.predict(state)

    def multistep_predict(self, state: State, n_steps: int) -> State:
        return self._core.multistep_predict(state, n_steps=n_steps)

    def update(self, state: State, measurement: np.ndarray) -> State:
        measurement = torch.from_numpy(measurement)
        return self._core.update(state, measurement)

    def missing(self, state: State) -> State:
        return self._core.missing(state)

    def project(self, state: State) -> Tuple[np.ndarray, np.ndarray]:
        mean, cov = self._core.project(state)
        mean, cov = mean.cpu().numpy(), cov.cpu().numpy()
        return mean, cov

    def singlestep_to_multistep_state(self, state: State) -> State:
        return self.singlestep_to_multistep_state(state)



"""
Custom collection support.
"""
from collections import defaultdict
from typing import Callable, List, Any, Dict


def nesteddict() -> defaultdict:
    """
    Creates nested dict (like defaultdict(dict) but recursive)

    Returns:
        nested defaultdict(dict)
    """
    return defaultdict(nesteddict)


def defaultdict_to_dict(nested: defaultdict) -> dict:
    """
    Converts defaultdict to dict (recursively).

    Motivation: defaultdict is useful for construction but might give some side effects when performing queries.

    Args:
        nested: Dictionary to convert

    Returns:
        Converted dictionary
    """
    if not isinstance(nested, defaultdict):
        return nested

    return {k: defaultdict_to_dict(v) for k, v in nested.items()}


def group_by(items: List[Any], func: Callable) -> Dict[Any, List[Any]]:
    """
    Groups list items into buckets (list) by custom function.

    Args:
        items: List of items that need to be grouped into "buckets"
        func: Grouping function (can be seen as bucket hash)

    Returns:
        Mapping (bucket_hash -> bucket_of_items)
    """
    buckets = defaultdict(list)

    for item in items:
        item_hash = func(item)
        buckets[item_hash].append(item)

    return dict(buckets)


def unpack_n(items: list, n: int) -> List[list]:
    """
    Unpacks list of tuples with n items into n lists.
    Handles empty list edge case.

    Args:
        items: List of tuples
        n: Number of items per tuples

    Returns:
        List of unpacked lists.
    """
    if len(items) == 0:
        return n * [[]]

    return [list(x) for x in zip(*items)]




"""
DynamicCatalog allows dynamic extensions of the factory methods.
"""
from typing import Callable, Type, List


class DynamicCatalog:
    """
    Dynamic catalog that is meant to be used with factory methods.
    Allows dynamic extension of the package factory method.
    """
    def __init__(self):
        self._catalog = {}

    @property
    def keys(self) -> List[str]:
        """
        Returns:
            List of catalog keys
        """
        return list(self._catalog.keys())

    def register(self, key: str) -> Callable[[Type], Type]:
        """
        Registers class to a catalog.

        Args:
            key: Class name

        Returns:
            Class
        """
        key = key.lower()

        if key in self._catalog:
            raise KeyError(f'Key "{key}" already registered!')

        def cls_register(cls: Type) -> Type:
            self._catalog[key] = cls
            return cls

        return cls_register

    def __getitem__(self, key: str) -> Type:
        key = key.lower()

        if key not in self._catalog:
            raise KeyError(f'Invalid key "{key}". Registered: {self.keys}.')

        return self._catalog[key]



"""
Color palette.
"""
from typing import Tuple

import matplotlib.colors as mcolors

ColorType = Tuple[int, int, int]

# BGR
RED = (0, 0, 255)
BLUE = (255, 0, 0)
GREEN = (0, 128, 0)
ORANGE = (0, 185, 255)
PINK = (147, 20, 255)
YELLOW = (0, 255, 255)
MAGENTA = (255, 0, 255)
WHITE = (255, 255, 255)
CYAN = (255, 255, 0)
DARK_GREEN = (0, 204, 0)
PURPLE = (51, 0, 51)
BLACK = (0, 0, 0)
GRAY = (128, 128, 128)
TEAL = (128, 128, 0)
OLIVE = (0, 128, 128)
NAVY = (128, 0, 0)
LIME = (0, 255, 0)
AQUA = (212, 255, 127)
SKY_BLUE = (235, 206, 136)
INDIGO = (130, 0, 75)
BROWN = (19, 69, 139)
TURQUOISE = (208, 224, 64)
CORAL = (80, 127, 255)
TAN = (140, 180, 210)
VIOLET = (238, 130, 238)


ALL_COLORS_EXPECT_BLACK = [
    RED, BLUE, GREEN, ORANGE, PINK,
    YELLOW, MAGENTA, WHITE, CYAN, DARK_GREEN,
    PURPLE, GRAY, TEAL, OLIVE,
    NAVY, LIME, SKY_BLUE, INDIGO,
    BROWN, TURQUOISE, CORAL, TAN, VIOLET
]


ALL_COLORS = ALL_COLORS_EXPECT_BLACK + [BLACK]


MATPLOTLIB_COLORS = list(mcolors.BASE_COLORS) + list(mcolors.CSS4_COLORS)



"""
Set of numpy functions applied on bounding box coordinates.
"""
import numpy as np


def affine_transform(warp: np.ndarray, bbox_coords: np.ndarray) -> np.ndarray:
    """
    Applied affine transform to bbox coordinates in format (xyxy or xywh).

    Args:
        warp: Warp matrix
        bbox_coords: Bounding box coordinates

    Returns:
        Warped bounding box coordinates
    """
    assert bbox_coords.shape == (4,)
    L = np.kron(np.eye(2, dtype=np.float32), warp[:, :2])
    T = np.concatenate([warp[:, 2], warp[:, 2]])
    t_bbox_coords = L @ bbox_coords + T
    return t_bbox_coords



"""
Utility functions for association algorithm.
"""
from motrack.tracker.tracklet import TrackletHistoryType


def filter_observations(history: TrackletHistoryType) -> TrackletHistoryType:
    """
    Remove all tracklet estimations. Estimations take place with
    last observation index if the observation is missing (missed detection).

    Disclaimer: This is coupled with MotionReId logic.

    Args:
        history:

    Returns:
        Tracklet observation history
    """
    last_frame_index: int = -1

    observation_history: TrackletHistoryType = []
    for index, bbox in history:
        if last_frame_index == index:
            continue
        last_frame_index = index
        observation_history.append((index, bbox))

    return observation_history



"""
Filter factory method.
Use `FILTER_CATALOG.register` to extend supported filter algorithms.
"""
from motrack.filter.algorithms.base import StateModelFilter
# noinspection PyUnresolvedReferences
from motrack.filter.algorithms.end_to_end import MotrackMotionFilterEndToEnd  # pylint: disable=unused-import
# noinspection PyUnresolvedReferences
from motrack.filter.algorithms.kalman_filter import BotSortKalmanWrapFilter  # pylint: disable=unused-import
# noinspection PyUnresolvedReferences
from motrack.filter.algorithms.no_motion import NoMotionFilter  # pylint: disable=unused-import
from motrack.filter.catalog import FILTER_CATALOG


def filter_factory(name: str, params: dict) -> StateModelFilter:
    """
    Filter (StateModel) factory.

    Args:
        name: Filter name (type)
        params: Filter creation parameters

    Returns:
        Created filter object
    """
    return FILTER_CATALOG[name](**params)



"""
Filter interface.
"""
from motrack.filter.algorithms.base import StateModelFilter
from motrack.filter.algorithms.kalman_filter import BotSortKalmanWrapFilter
from motrack.filter.algorithms.no_motion import NoMotionFilter
from motrack.filter.factory import filter_factory



"""
Tracker inference. Output can directly be evaluated using the TrackEval repo.
"""
import logging
import os
import shutil
from dataclasses import asdict

import hydra
import yaml
from omegaconf import DictConfig

from motrack.common.project import CONFIGS_PATH
from motrack.config_parser import GlobalConfig
from motrack.datasets import dataset_factory
from motrack.object_detection import DetectionManager
from motrack.tracker import tracker_factory
from motrack.utils import pipeline
from motrack.tools import run_tracker_inference, run_tracker_postprocess

logger = logging.getLogger('TrackerInference')


@pipeline.task('inference')
def run_inference(cfg: GlobalConfig) -> None:
    if os.path.exists(cfg.experiment_path):
        if cfg.eval.override:
            user_input = 'yes'  # `input` from config
        else:
            user_input = input(f'Experiment on path "{cfg.experiment_path}" already exists. '
                               f'Are you sure you want to override it? [yes/no] ').lower()
        if user_input in ['yes', 'y']:
            shutil.rmtree(cfg.experiment_path)
        else:
            logger.info('Aborting!')
            return

    tracker_active_output = os.path.join(cfg.experiment_path, 'active')
    tracker_all_output = os.path.join(cfg.experiment_path, 'all')

    logger.info(f'Saving tracker inference on path "{cfg.experiment_path}".')

    dataset = dataset_factory(
        dataset_type=cfg.dataset.type,
        path=cfg.dataset.fullpath,
        params=cfg.dataset.params,
        test=cfg.eval.split == 'test'
    )

    detection_manager = DetectionManager(
        inference_name=cfg.object_detection.type,
        inference_params=cfg.object_detection.params,
        lookup=cfg.object_detection.load_lookup() if cfg.object_detection.lookup_path is not None else None,
        dataset=dataset,
        cache_path=cfg.object_detection.cache_path,
        oracle=cfg.object_detection.oracle
    )

    tracker = tracker_factory(
        name=cfg.algorithm.name,
        params=cfg.algorithm.params
    )

    run_tracker_inference(
        dataset=dataset,
        tracker=tracker,
        detection_manager=detection_manager,
        tracker_active_output=tracker_active_output,
        tracker_all_output=tracker_all_output,
        clip=cfg.dataset.type != 'MOT17',
        scene_pattern=cfg.dataset_filter.scene_pattern,
        load_image=cfg.eval.load_image
    )

    # Save tracker config
    tracker_config_path = os.path.join(cfg.experiment_path, 'config.yaml')
    with open(tracker_config_path, 'w', encoding='utf-8') as f:
        yaml.safe_dump(asdict(cfg), f)

    if cfg.eval.postprocess:
        logger.info('Performing inference postprocessing...')
        tracker_postprocess_output = os.path.join(cfg.experiment_path, 'postprocess')
        run_tracker_postprocess(
            dataset=dataset,
            tracker_active_output=tracker_active_output,
            tracker_all_output=tracker_all_output,
            tracker_postprocess_output=tracker_postprocess_output,
            postprocess_cfg=cfg.postprocess,
            scene_pattern=cfg.dataset_filter.scene_pattern
        )


@hydra.main(config_path=CONFIGS_PATH, config_name='movesort', version_base='1.1')
def main(cfg: DictConfig):
    # noinspection PyTypeChecker
    run_inference(cfg)


if __name__ == '__main__':
    main()



"""
YOLOX inference class.
"""
from typing import List, Optional, Tuple, Union, Any

import numpy as np

from motrack.library.cv.bbox import LabelType
from motrack.object_detection.algorithms.base import ObjectDetectionInference
from motrack.object_detection.catalog import OBJECT_DETECTION_CATALOG
from motrack.utils.lookup import LookupTable


@OBJECT_DETECTION_CATALOG.register('yolox')
class YOLOXInference(ObjectDetectionInference):
    """
    Object Detection - YOLOX
    """
    def __init__(
        self,
        model_path: str,
        accelerator: str,
        conf: float = 0.01,
        min_bbox_area: int = 0,
        legacy: bool = True,
        lookup: Optional[LookupTable] = None
    ):
        super().__init__(lookup=lookup)
        try:
            from motrack.object_detection.yolox import YOLOXPredictor
        except ImportError as e:
            raise ImportError('Please install PyTorch from "https://pytorch.org/"'
                              'and YOLOX package from "https://github.com/Megvii-BaseDetection/YOLOX"!') from e

        self._yolox = YOLOXPredictor(
            checkpoint_path=model_path,
            accelerator=accelerator,
            conf_threshold=conf,
            legacy=legacy
        )
        self._min_bbox_area = min_bbox_area

    def predict_raw(self, image: np.ndarray) -> Any:
        output, _ = self._yolox.predict(image)
        return output

    def postprocess(self, image: np.ndarray, raw: Any) -> Tuple[Union[np.ndarray, List[float]], List[LabelType], Union[np.ndarray, list]]:
        h, w, _ = image.shape

        # Filter small bboxes
        bboxes = raw[:, :4]
        areas = (bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])
        output = raw[areas >= self._min_bbox_area]

        # Process bboxes
        bboxes = output[:, :4]
        bboxes[:, [0, 2]] /= w
        bboxes[:, [1, 3]] /= h

        # Process classes
        classes = output[:, 6]

        # Process confidences
        confidences = output[:, 4]

        return bboxes, classes, confidences



"""
Inference support for FastReId framework.
Reference: https://github.com/JDAI-CV/fast-reid
"""
import logging
from typing import Optional

import cv2
import numpy as np

from motrack.reid.algorithms.base import BaseReID
from motrack.reid.catalog import REID_CATALOG

logger = logging.getLogger('FastReIDOnnx')


@REID_CATALOG.register('fastreid-onnx')
class FastReIDOnnx(BaseReID):
    """
    FastReIDOnnx framework support for any models exported in ONNX format.
    """
    def __init__(
        self,
        model_path: str,
        height: int = 256,
        width: int = 256,
        cache_path: Optional[str] = None,
        batch_inference: bool = False
    ):
        """
        Args:
            model_path: Path where the ONNX model export is stored
            height: Image input height
            width: Image input width
            cache_path: Inference cache path
            batch_inference: Use batch inference mode if supported by the ONNX export
                - This makes inference faster
        """
        super().__init__(cache_path=cache_path, batch_inference=batch_inference)

        import onnxruntime
        logger.info(f'Runtime device: {onnxruntime.get_device()}.')
        self._ort_session = onnxruntime.InferenceSession(model_path)
        self._input_name = self._ort_session.get_inputs()[0].name
        self._height = height
        self._width = width

    def preprocess(self, image: np.ndarray) -> np.ndarray:
        # the model expects RGB inputs
        image = image[:, :, ::-1]

        # Apply pre-processing to frame.
        image = cv2.resize(image, (self._width, self._height), interpolation=cv2.INTER_CUBIC)
        image = image.astype(np.float32).transpose(2, 0, 1)  # (1, 3, h, w)
        return image

    def postprocess(self, features: np.ndarray) -> np.ndarray:
        norm = np.linalg.norm(features, ord=2, axis=1, keepdims=True)
        return features / (norm + np.finfo(np.float32).eps)

    def inference(self, image: np.ndarray) -> np.ndarray:
        return self._ort_session.run(None, {self._input_name: image})[0]



"""
Interface for re-identification appearance extractor
"""
import os
from abc import ABC, abstractmethod
from typing import List, Optional
import copy

import numpy as np

from motrack.library.cv.bbox import BBox
from motrack.library.numpy_utils.io import load_npy, store_npy


class BaseReID(ABC):
    """.
    Interface for re-identification appearance extractor
    """
    def __init__(self, cache_path: Optional[str] = None, batch_inference: bool = False):
        """
        Args:
            cache_path: Inference cache path
            batch_inference: Use batch inference mode if supported by the ONNX export
                - This makes inference faster
        """
        self._cache_path = cache_path
        self._batch_inference = batch_inference

    @abstractmethod
    def preprocess(self, image: np.ndarray) -> np.ndarray:
        """
        Preprocesses the input frame. Ref: https://github.com/JDAI-CV/fast-reid/blob/master/tools/deploy/onnx_inference.py

        Args:
            image: Object crop

        Returns:
            frame features (not normalized)
        """

    @abstractmethod
    def postprocess(self, features: np.ndarray) -> np.ndarray:
        """
        Normalized the input features to L2 ball.

        Args:
            features: Un-normalized features

        Returns:
            Normalized features
        """

    @abstractmethod
    def inference(self, image: np.ndarray) -> np.ndarray:
        """
        Extracts object appearance features for given frame.

        Args:
            image: Raw frame

        Returns:
            Object frame appearance features
        """

    def extract_features(self, image: np.ndarray) -> np.ndarray:
        data = self.preprocess(image)
        features = self.inference(data)
        return self.postprocess(features)

    def extract_objects_features(self, frame: np.ndarray, bboxes: List[BBox], frame_index: int, scene: Optional[str] = None) -> np.ndarray:
        """
        Performs object appearance features for given frame and list of objects' bounding boxes.

        Args:
            frame: Raw frame
            bboxes: Detected bounding boxes
            frame_index: Frame number (index)
            scene: Scene name (optional, for caching)

        Returns:
            Array of object appearance features for each detected object
        """
        if self._cache_path is not None:
            # Load features from cache instead of running the inference
            assert scene is not None, 'Scene must be specified in order to perform caching!'
            scene_frame_cache_path = self._get_scene_frame_cache_path(scene, frame_index)
            if os.path.exists(scene_frame_cache_path):
                return load_npy(scene_frame_cache_path)


        data_list: List[np.ndarray] = []

        # Pack batch input
        for bbox in bboxes:
            # Clip bbox to image
            bbox = copy.deepcopy(bbox)
            bbox.clip()
            assert bbox.area > 0, 'Bounding box has no area!'

            crop = bbox.crop(frame)
            data = self.preprocess(crop)
            data_list.append(data)

        if len(data_list) == 0:
            objects_features = np.empty(0, dtype=np.float32)
        else:
            data = np.stack(data_list, 0)

            if self._batch_inference:
                objects_features = self.inference(data)
            else:
                objects_features = np.concatenate([self.inference(data[i][np.newaxis]) for i in range(data.shape[0])], 0)

            objects_features = self.postprocess(objects_features)

        if self._cache_path is not None:
            # Store features to cache for faster future inference
            assert scene is not None, 'Scene must be specified in order to perform caching!'
            scene_frame_cache_path = self._get_scene_frame_cache_path(scene, frame_index)
            store_npy(objects_features, scene_frame_cache_path)

        return objects_features


    def _get_scene_frame_cache_path(self, scene: str, frame_index: int) -> str:
        """
        Args:
            scene: Scene name
            frame_index: Scene frame number

        Returns:
            Path where the ReId data is cached.
        """
        return os.path.join(self._cache_path, scene, f'reid_{frame_index:06d}.npy')



"""
Load GMC results from file - pre-calculated.
"""
import os
from typing import Optional, Dict

import numpy as np

from motrack.cmc.algorithms.base import CameraMotionCompensation
from motrack.cmc.catalog import CMC_CATALOG


@CMC_CATALOG.register('gmc-from-file')
class GMCFromFile(CameraMotionCompensation):
    """
    Loads precalculated GMC warps from a directory for each scene.
    """
    LINE_SEP = '\t'

    def __init__(self, dirpath: str):
        """
        Args:
            dirpath: Path to directory where precalculated GMC warps are stored.
        """
        self._gmc_lookup = self._parse_gmc_directory(dirpath)

    @staticmethod
    def _get_gmc_filename(scene: str) -> str:
        """
        Gets GMC filename based on the scene name.

        Args:
            scene: Scene name

        Returns:

        """
        filename = f'GMC-{scene}.txt'

        # DanceTrack naming (special case)
        if scene.startswith('dancetrack'):
            filename = filename.replace('dancetrack', 'dancetrack-')

        return filename

    @staticmethod
    def _parse_gmc_directory(path: str) -> Dict[str, np.ndarray]:
        """
        Parses all files in GMC directory.

        Args:
            path: Path to GMC files.

        Returns:
            GMC warp lookup
        """
        gmc_lookup = {}

        files = os.listdir(path)
        for file in files:
            filepath = os.path.join(path, file)
            with open(filepath, 'r', encoding='utf-8') as f:
                lines = [line for line in f.readlines() if len(line) > 0]

            n_lines = len(lines)
            warps = np.zeros(shape=(n_lines, 2, 3), dtype=np.float32)
            for line_i, line in enumerate(lines):
                tokens = line.strip().split(GMCFromFile.LINE_SEP)[1:]
                for token_i in range(len(tokens)):
                    r, c = token_i // 3, token_i % 3
                    warps[line_i, r, c] = float(tokens[token_i])

            gmc_lookup[file] = warps

        return gmc_lookup


    def apply(self, frame: np.ndarray, frame_index: int, scene: Optional[str] = None) -> np.ndarray:
        assert scene is not None, 'Scene name is required in order to load GMC warps from a file!'
        scene_file = self._get_gmc_filename(scene)
        warp = self._gmc_lookup[scene_file][frame_index, :, :]
        warp[0, 2] /= frame.shape[1]
        warp[1, 2] /= frame.shape[0]
        return warp



"""
TrackEval inference IO support.
"""
import copy
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TextIO, Dict, Optional

from motrack.library.cv.bbox import PredBBox, BBox
from motrack.tracker.tracklet import Tracklet

TRACKER_INFERENCE_HEADER = [
    'frame_id', 'id',
    'bbox_left', 'bbox_top', 'bbox_width', 'bbox_height', 'conf',
    '-1_0', '-1_1', '-1_2'
]


class TrackerInferenceWriter:
    """
    Writes tracker inference in MOT format. Can be used afterward for evaluation.
    (RAII)
    """
    def __init__(self, output_path: str, scene_name: str, image_height: int, image_width: int, clip: bool = True):
        """
        Args:
            output_path: Tracker inference output directory path
            scene_name: Scene name
            image_height: Image height (required for coordinates up-scaling)
            image_width: Image height (required for coordinates up-scaling)
            clip: Clip tracklet bbox
        """
        self._output_path = output_path
        self._scene_name = scene_name
        self._scene_output_path = os.path.join(output_path, f'{scene_name}.txt')
        self._image_height = image_height
        self._image_width = image_width
        self._clip = clip

        # State
        self._writer = None

    def open(self) -> None:
        """
        Opens writer.
        """
        Path(self._output_path).mkdir(parents=True, exist_ok=True)
        self._writer: TextIO = open(self._scene_output_path, 'w', encoding='utf-8')  # pylint: disable=consider-using-with

    def close(self) -> None:
        """
        Closes writer.
        """
        if self._writer is not None:
            self._writer.close()

    def write(self, frame_index: int, tracklet: Tracklet) -> None:
        """
        Writes info about one tracker tracklet.
        One tracklet - one row.

        Args:
            frame_index: Current frame index
            tracklet: Tracklet
        """
        bbox = tracklet.bbox
        if self._clip:
            bbox = copy.deepcopy(bbox)
            bbox.clip()

        left = round(bbox.upper_left.x * self._image_width)
        top = round(bbox.upper_left.y * self._image_height)
        width = round(bbox.width * self._image_width)
        height = round(bbox.height * self._image_height)

        cells = [
            str(frame_index + 1), str(tracklet.id),
            str(left), str(top),
            str(width), str(height),
            str(bbox.conf),
            '-1', '-1', '-1'
        ]

        row = ','.join(cells)
        self._writer.write(f'{row}\n')

    def __enter__(self) -> 'TrackerInferenceWriter':
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@dataclass
class TrackerInferenceRead:
    frame_index: int
    frame_id: str
    objects: Dict[str, PredBBox]


class TrackerInferenceReader:
    """
    Reads tracker inference in MOT format.
    (RAII)
    """
    BBOX_LABEL = 'Person'

    def __init__(self, output_path: str, scene_name: str, image_height: int, image_width: int):
        """
        Args:
            output_path: Tracker inference output directory path
            scene_name: Scene name
            image_height: Image height (required for coordinates up-scaling)
            image_width: Image height (required for coordinates up-scaling)
        """
        self._output_path = output_path
        self._scene_name = scene_name
        self._scene_output_path = os.path.join(output_path, f'{scene_name}.txt')
        self._image_height = image_height
        self._image_width = image_width

        # State
        self._reader = None
        self._curr_frame_id = None
        self._last_line = None

    def open(self) -> None:
        """
        Opens reader.
        """
        self._reader: TextIO = open(self._scene_output_path, 'r', encoding='utf-8')  # pylint: disable=consider-using-with

    def close(self) -> None:
        """
        Closes reader.
        """
        if self._reader is not None:
            self._reader.close()

    def read(self) -> Optional[TrackerInferenceRead]:
        """
        Reads information about next frame.

        Returns:
            Dictionary with fields:
            - `frame_index` for frame index
            - `frame_id` for frame id (frame_index + 1)
            - `objects` all objects present in the current frame
        """
        line = self._reader.readline() if self._last_line is None else self._last_line
        if not line:
            # End of file
            return None
        self._last_line = None

        data = None
        while True:
            frame_id, tracklet_id, bbox_left, bbox_top, bbox_width, bbox_height, conf, _, _, _ = line.split(',')
            frame_index = int(frame_id) - 1
            if self._curr_frame_id is None or self._curr_frame_id == frame_id:
                if self._curr_frame_id is None:
                    # Create data object
                    data = TrackerInferenceRead(
                        frame_index=frame_index,
                        frame_id=frame_id,
                        objects={}
                    )

                    self._curr_frame_id = frame_id

                # Save object
                bbox_left, bbox_top, bbox_width, bbox_height, conf = [float(v) for v in [bbox_left, bbox_top, bbox_width, bbox_height, conf]]
                bbox = PredBBox.create(
                    bbox=BBox.from_xywh(
                        bbox_left / self._image_width,
                        bbox_top / self._image_height,
                        bbox_width / self._image_width,
                        bbox_height / self._image_height
                    ),
                    label=self.BBOX_LABEL,  # Note: This is specialized for MOT20 format
                    conf=conf
                )
                data.objects[tracklet_id] = bbox

                # Read next line
                line = self._reader.readline()
                if not line:
                    # End of file
                    self._curr_frame_id = None
                    self._last_line = None
                    return data
            else:
                self._curr_frame_id = None
                self._last_line = line
                return data

    def __enter__(self) -> 'TrackerInferenceReader':
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Tracklet Dataclass.
"""
import enum
from multiprocessing import Value, Lock
from typing import Tuple, ClassVar, Optional, List, Any, Union

from motrack.library.cv.bbox import PredBBox

_TRACKLET_DEFAULT_MAX_HISTORY = 32

TrackletHistoryType = List[Tuple[int, PredBBox]]


# Conventions
class TrackletCommonData(enum.Enum):
    """
    Tracklet common data type used.
    """
    APPEARANCE = 'appearance'
    APPEARANCE_BUFFER = 'appearance-buffer'


class TrackletState(enum.Enum):
    """
    Tracklet state:
    - NEW: New detected object (not confirmed)
    - ACTIVE: Active tracklet (confirmed)
    - LOST: Inactive tracklet (lost due occlusion, etc.)
    """
    NEW = enum.auto()
    ACTIVE = enum.auto()
    LOST = enum.auto()
    DELETED = enum.auto()


class Tracklet:
    """
    Tracking object state
    """
    Counter: ClassVar[Value] = Value('i', 0)
    CounterLock: ClassVar[Lock] = Lock()

    def __init__(
        self,
        bbox: PredBBox,
        frame_index: int,
        max_history: int = _TRACKLET_DEFAULT_MAX_HISTORY,
        _id: Optional[int] = None,
        state: TrackletState = TrackletState.NEW
    ):
        """
        Initializes new tracklet.

        Args:
            bbox: Initial bbox
            frame_index: Frame index when tracklet was created
            max_history: Max tracklet history (old data is deleted)
            _id: Set tracklet if explicitly
            state: Initial state

        Returns:
            Initialized tracklet
        """
        if _id is None:
            with Tracklet.CounterLock:
                Tracklet.Counter.value += 1
                self._id = Tracklet.Counter.value
        else:
            self._id = _id

        # Update bbox
        bbox.id = self._id

        self._max_history = max_history
        self._start_frame_index = frame_index

        # State
        self._history: TrackletHistoryType = [(frame_index, bbox)]
        self._total_matches = 1
        self._total_lost = 0
        self._state = state
        self._data = None

    def __repr__(self) -> str:
        return f'Tracklet(id={self._id}, bbox={self.bbox}, state={self._state})'

    def __hash__(self) -> int:
        return self._id

    @property
    def state(self) -> TrackletState:
        """
        Gets tracklet state.

        Returns:
            Tracklets state.
        """
        return self._state

    @state.setter
    def state(self, new_state: TrackletState) -> None:
        """
        Sets tracklet state.

        Args:
            new_state: New tracklet state
        """
        self._state = new_state

    @property
    def is_tracked(self) -> bool:
        """
        Checks if tracklet is being tracked.

        Returns:
            True if tracklets is tracked else False
        """
        return self._state in [TrackletState.ACTIVE, TrackletState.LOST]

    @property
    def start_frame_index(self) -> int:
        """
        Gets frame when the Tracklet was initialized.

        Returns:
            Tracklet first frame.
        """
        return self._start_frame_index

    @property
    def age(self) -> int:
        """
        Gets tracklet age.

        Returns:
            Tracklet age
        """
        return self.frame_index - self._start_frame_index

    @property
    def lost_time(self) -> int:
        """
        Gets number of frames that tracklet has LOST state

        Example: (match, missing, match, missing, missing) -> returns 2

        Returns:
            Number of unmatched frames
        """
        return self._total_lost

    @property
    def id(self) -> int:
        """
        Gets tracklet id.

        Returns:
            Tracklet id
        """
        return self._id


    @property
    def latest(self) -> Tuple[int, PredBBox]:
        """
        The latest detection of one tracklet.

        Returns:
            id and PredBBox
        """
        return self._history[-1]

    @property
    def first(self) -> Tuple[int, PredBBox]:
        """
        First detection (last detection in history) of one tracklet.

        Returns:
            id and PredBBox
        """
        return self._history[0]

    @property
    def history(self) -> TrackletHistoryType:
        """
        Gets full tracklet history.

        Returns:
            Tracklet history
        """
        return self._history

    @property
    def bbox(self) -> PredBBox:
        """
        Gets current tracklet bbox.

        Returns:
            Current bbox
        """
        return self.latest[1]

    @property
    def frame_index(self) -> int:
        """
        Gets current (last) frame index.

        Returns:
            Frame index
        """
        return self.latest[0]

    @property
    def total_matches(self) -> int:
        """
        Total number of tracklet matches with detected object.

        Returns:
            Total number of tracklet matches
        """
        return self._total_matches

    def __len__(self) -> int:
        """
        Returns:
            tracklet length
        """
        return len(self._history)

    def update(self, bbox: PredBBox, frame_index: int, state: Optional[TrackletState] = None) -> 'Tracklet':
        """
        Updates tracklet BBox (and history).
        Disclaimer: Updates bbox id!

        Args:
            bbox: New bbox
            frame_index: Current frame index
            state: New tracklet state
        """
        bbox.id = self._id  # Update bbox id

        self._history.append((frame_index, bbox))

        # Remove `obsolete` history
        while len(self._history) > self._max_history:
            self._history.pop(0)

        if state is not None:
            self._state = state

        # Update state stats
        if self._state in [TrackletState.NEW, TrackletState.ACTIVE]:
            self._total_matches += 1
            self._total_lost = 0
        elif self._state == TrackletState.LOST:
            self._total_lost += 1

        return self

    def set(self, key: Union[str, TrackletCommonData], value: Any) -> None:
        """
        Updates Tracklet key-value data.

        Args:
            key: Data key
            value: Data value
        """
        if self._data is None:
            self._data = {}
        self._data[key] = value

    def get(self, key: Union[str, TrackletCommonData]) -> Any:
        """
        Gets Tracklet data by key.

        Args:
            key: Data key

        Returns:
            Data value if key is in data else None
        """
        if self._data is None or key not in self._data:
            return None

        return self._data[key]



"""
Tracker config for tools scripts.
"""
import json
import os
from dataclasses import dataclass, field
from typing import Optional

from hydra.core.config_store import ConfigStore

from motrack.common import project
from motrack.utils.lookup import LookupTable


@dataclass
class PathConfig:
    """
    Path configs
    - master: location where all final and intermediate results are stored
    - assets: location where datasets can be found
    """
    master: str = project.MASTER_PATH
    assets: str = project.ASSETS_PATH

    @classmethod
    def default(cls) -> 'PathConfig':
        """
        Default path configuration is used if it is not defined in configs.

        Returns: Path configuration.
        """
        return cls(
            master=project.OUTPUTS_PATH,
            assets=project.ASSETS_PATH
        )


@dataclass
class DatasetConfig:
    """
    Dataset config.
    - name: Dataset name
    - path: Path to the dataset
    - params: Ctor parameters
    """
    type: str
    path: str
    params: dict = field(default_factory=dict)

    def __post_init__(self):
        """
        Fullpath is resolved from global config (path is required)
        """
        self.fullpath = None


@dataclass
class FilterConfig:
    """
    Filter (motion model) config.
    """
    type: str
    params: dict


@dataclass
class ObjectDetectionInferenceConfig:
    """
    Object detection config.
    - type: Architecture
    - params: Ctor params
    - lookup_path: Class mappings saved as a json file (in LookupTable format)
    - cache_path: Use cache for faster inference
    - oracle: Use GT detections instead running inference
    """
    type: str
    params: dict
    lookup_path: Optional[str] = None
    cache_path: Optional[str] = None
    oracle: bool = False

    def load_lookup(self) -> LookupTable:
        """
        Loads lookup file.

        Returns:
            Loaded lookup
        """
        with open(self.lookup_path, 'r', encoding='utf-8') as f:
            return LookupTable.deserialize(json.load(f))


@dataclass
class TrackerAlgorithmConfig:
    name: str
    params: dict
    requires_image: bool = False


@dataclass
class TrackerVisualizeConfig:
    fps: int = 20
    new_object_length: int = 5
    option: str = 'all'

    def __post_init__(self) -> None:
        """
        Validation.
        """
        options = ['active', 'all', 'postprocess']
        assert self.option in options, f'Invalid option "{self.option}". Available: {options}.'


@dataclass
class TrackerPostprocessConfig:
    init_threshold: int = 2  # Activate `init_threshold` starting bboxes
    linear_interpolation_threshold: int = 3  # Maximum distance to perform linear interpolation
    linear_interpolation_min_tracklet_length: int = 30  # Minimum tracklet length to perform linear interpolation
    min_tracklet_length: int = 20  # Remove all tracklets that are shorter than this


@dataclass
class TrackerEvalConfig:
    split: str
    postprocess: bool = field(default=False)
    override: bool = field(default=False)
    load_image: bool = field(default=True)


@dataclass
class DatasetFilterConfig:
    scene_pattern: str = '(.*?)'  # All


@dataclass
class GlobalConfig:
    experiment: str
    dataset: DatasetConfig
    eval: TrackerEvalConfig
    object_detection: ObjectDetectionInferenceConfig
    algorithm: TrackerAlgorithmConfig
    dataset_filter: DatasetFilterConfig = field(default_factory=DatasetFilterConfig)
    path: PathConfig = field(default_factory=PathConfig)
    postprocess: TrackerPostprocessConfig = field(default_factory=TrackerPostprocessConfig)
    visualize: TrackerVisualizeConfig = field(default_factory=TrackerVisualizeConfig)

    @property
    def experiment_path(self) -> str:
        """
        Path where tracker results can be found.
        """
        return os.path.join(self.path.master, self.dataset.type, self.algorithm.name, self.experiment, self.eval.split)

    def __post_init__(self):
        """
        Postprocess.
        """
        self.dataset.fullpath = os.path.join(self.path.assets, self.dataset.path, self.eval.split)


# Configuring hydra config store
# If config has `- global_config` in defaults then
# full config is recursively instantiated
cs = ConfigStore.instance()
cs.store(name='global_config', node=GlobalConfig)



"""
BBox implementation
"""
from dataclasses import dataclass, field
from typing import Optional, List, Tuple, Union

import cv2
import numpy as np

from motrack.library.numpy_utils.bbox import affine_transform


@dataclass
class Point:
    """
    Point in 2D (x, y).
    """
    x: float
    y: float

    def copy(self) -> 'Point':
        """
        Returns: Copies object
        """
        return Point(x=self.x, y=self.y)

    def clip(self) -> None:
        """
        Clips coords to [0, 1] range.
        """
        self.x = max(0.0, min(1.0, self.x))
        self.y = max(0.0, min(1.0, self.y))

    def __lt__(self, other: 'Point') -> bool:
        return self.x < other.x and self.y < other.y

    def __le__(self, other: 'Point') -> bool:
        return self.x <= other.x and self.y <= other.y

    def __eq__(self, other: 'Point') -> bool:
        return all([np.isclose(self.x, other.x), np.isclose(self.y, other.y)])

    def as_numpy_xy(self, dtype: np.dtype = np.float32) -> np.ndarray:
        """
        Converts point to numpy array.

        Args:
            dtype: Numpy dtype

        Returns:
            Numpy point
        """
        return np.array([self.x, self.y], dtype=dtype)


@dataclass
class BBox:
    """
    BBox (rectangle) defined by two upper left and bottom right corners.
    """
    upper_left: Point
    bottom_right: Point

    def __post_init__(self):
        """
        Validation
        """
        return self.upper_left <= self.bottom_right

    def clip(self) -> None:
        """
        Clips coords to [0, 1] range.
        """
        self.bottom_right.clip()
        self.upper_left.clip()

    def copy(self) -> 'BBox':
        """
        Returns: Deepcopy of object
        """
        return BBox(
            upper_left=self.upper_left.copy(),
            bottom_right=self.bottom_right.copy()
        )

    @property
    def width(self) -> float:
        """
        Returns: Width
        """
        return self.bottom_right.x - self.upper_left.x

    @property
    def height(self) -> float:
        """
        Returns: Height
        """
        return self.bottom_right.y - self.upper_left.y

    @property
    def area(self) -> float:
        """
        Returns: Area
        """
        return self.height * self.width

    @property
    def center(self) -> Point:
        """
        Returns: Center
        """
        return Point(x=(self.bottom_right.x + self.upper_left.x) / 2, y=(self.bottom_right.y + self.upper_left.y) / 2)

    @property
    def xyxy(self) -> np.ndarray:
        """
        Returns:
            Numpy array in xyxy format
        """
        return np.array([self.upper_left.x, self.upper_left.y, self.bottom_right.x, self.bottom_right.y],
                        dtype=np.float32)

    def __eq__(self, other: 'BBox') -> bool:
        return self.upper_left == other.upper_left and self.bottom_right == other.bottom_right

    def intersection(self, other: 'BBox') -> Optional['BBox']:
        """
        Returns BBox if intersection exists else returns None

        Args:
            other: Other BBox

        Returns:
            Intersection of two bboxes
        """
        upper = max(self.upper_left.y, other.upper_left.y)
        bottom = min(self.bottom_right.y, other.bottom_right.y)
        left = max(self.upper_left.x, other.upper_left.x)
        right = min(self.bottom_right.x, other.bottom_right.x)

        if upper > bottom or left > right:
            return None

        return BBox(
            upper_left=Point(x=upper, y=left),
            bottom_right=Point(x=bottom, y=right)
        )

    def iou(self, other: 'BBox') -> float:
        """
        Intersection Over Union / Jaccard Index

        Args:
            other: Other BBox

        Returns: IOU
        """
        intersection = self.intersection(other)
        if intersection is None:
            return 0.0

        union_area = self.area + other.area - intersection.area
        if union_area == 0.0:
            return 0.0

        return intersection.area / union_area

    def max_iou(self, others: List['BBox']) -> Tuple[float, int]:
        """
        Max iou over all other bboxes

        Args:
            others: List of BBoxes

        Returns: Value, index
        """
        ious = [self.iou(other) for other in others]
        index = int(np.argmax(ious))
        return ious[index], index

    @classmethod
    def from_xyxy(cls, x1: float, y1: float, x2: float, y2: float, clip: bool = False) -> 'BBox':
        """
        Creates BBox from xyxy format

        Args:
            x1: x1
            y1: y1
            x2: x2
            y2: y2
            clip: Clip bbox coordinates to [0, 1] range

        Returns: Bbox
        """
        bbox = cls(
            upper_left=Point(x=x1, y=y1),
            bottom_right=Point(x=x2, y=y2),
        )

        if clip:
            bbox.clip()

        return bbox

    @classmethod
    def from_xywh(cls, x: float, y: float, h: float, w: float, clip: bool = False) -> 'BBox':
        """
        Creates BBox from xyhw format

        Args:
            x: left
            y: up
            h: height
            w: width
            clip: Clip bbox coordinates to [0, 1] range

        Returns: Bbox
        """
        x1, y1, x2, y2 = x, y, x + h, y + w
        return cls.from_xyxy(x1, y1, x2, y2, clip=clip)

    @classmethod
    def from_cxywh(cls, x: float, y: float, w: float, h: float, clip: bool = False) -> 'BBox':
        """
        Creates BBox from cxywh format

        Args:
            x: center x
            y: center y
            w: width
            h: height
            clip: Clip bbox coordinates to [0, 1] range

        Returns: Bbox
        """
        x1, y1, x2, y2 = x - w / 2, y - h / 2, x + w / 2, y + h / 2
        return cls.from_xyxy(x1, y1, x2, y2, clip=clip)

    def as_numpy_xyxy(self, dtype: np.dtype = np.float32) -> np.ndarray:
        """
        Converts Bbox to xyxy numpy array.

        Args:
            dtype: Numpy array dtype (default: np.float32)

        Returns:
            BBox xyxy coords as a numpy array.
        """
        return np.array([self.upper_left.x, self.upper_left.y, self.bottom_right.x, self.bottom_right.y], dtype=dtype)

    def as_numpy_xywh(self, dtype: np.dtype = np.float32) -> np.ndarray:
        """
        Converts Bbox to yxwh numpy array.

        Args:
            dtype: Numpy array dtype (default: np.float32)

        Returns:
            BBox yxwh coords as a numpy array.
        """
        return np.array([self.upper_left.x, self.upper_left.y, self.width, self.height], dtype=dtype)

    def as_numpy_cxywh(self, dtype: np.dtype = np.float32) -> np.ndarray:
        """
        Converts Bbox to cxyhw numpy array.

        Args:
            dtype: Numpy array dtype (default: np.float32)

        Returns:
            BBox cxyhw coords as a numpy array.
        """
        return np.array([*self.center, self.width, self.height], dtype=dtype)

    @staticmethod
    def clip_coords(x1: float, y1: float, x2: float, y2: float) -> Tuple[float, float, float, float]:
        """
        Clip coordinates to [0, 1] range such that x1 <= x2 and y1 <= y2.

        Args:
            x1: x1
            y1: y1
            x2: x2
            y2: y2

        Returns:
            Clipped coordinates
        """
        x1, y1, x2, y2 = [min(1.0, max(v, 0.0)) for v in [x1, y1, x2, y2]]
        x2 = max(x1, x2)  # assumption: x1 <= x2
        y2 = max(y1, y2)  # assumption: y1 <= y2

        return x1, y1, x2, y2

    def crop(self, image: np.ndarray) -> np.ndarray:
        """
        Crops image

        Args:
            image: Raw image

        Returns: Crop
        """
        x1, y1, x2, y2 = self.scaled_xyxy_from_image(image)
        return image[y1:y2, x1:x2, :].copy()

    def scaled_xyxy_from_image(self, image: np.ndarray) -> Tuple[int, int, int, int]:
        """
        Scales coordinates to given image

        Args:
            image: Image

        Returns: coord in xyxy format (scaled)
        """
        h, w = image.shape[:2]
        return round(self.upper_left.x * w), round(self.upper_left.y * h), \
            round(self.bottom_right.x * w), round(self.bottom_right.y * h)

    def draw(self, image: np.ndarray, color: Tuple[int, int, int] = (0, 0, 255), thickness: int = 2) -> np.ndarray:
        """
        Draws bbox on image

        Args:
            image: Image
            color: BBox color
            thickness: BBox thickness

        Returns: Image with drawn bbox
        """
        y1, x1, y2, x2 = self.scaled_xyxy_from_image(image)
        # noinspection PyUnresolvedReferences
        return cv2.rectangle(image, (y1, x1), (y2, x2), color, thickness)

    def affine_transform(self, warp: np.ndarray) -> 'BBox':
        """
        Performs affine transformation on the bbox (creates new bbox)

        Args:
            warp: Warp matrix 2x3

        Returns:
            New transformed bbox
        """
        points = self.as_numpy_xyxy()
        t_points = affine_transform(warp, points)
        return BBox.from_xyxy(*t_points)



LabelType = Union[int, str]


@dataclass
class PredBBox(BBox):
    """
    BBox with class label and detection confidence (optional).
    """
    label: LabelType = -1
    conf: Optional[float] = field(default=None)

    @property
    def aspect_ratio(self) -> float:
        return self.width / self.height

    @property
    def conf_annot(self) -> str:
        """
        Returns:
            Confidence annotation (as string).
        """
        return f'{100*self.conf:.1f}%' if self.conf is not None else 'GT'

    @property
    def compact_repr(self) -> str:
        """
        Returns: Compact PredBBox repr
        """
        coords_str = f'[{self.upper_left.x}, {self.upper_left.y}, {self.bottom_right.x}, {self.bottom_right.y}]'
        return f'PredBBox({coords_str} {self.label} ({self.conf_annot}))'

    @classmethod
    def create(cls, bbox: BBox, label: LabelType, conf: Optional[float] = None):
        """
        Creates PredBbox from regular BBox

        Args:
            bbox: BBox
            label: Label
            conf: Confidence

        Returns:
            Prediction BBox
        """
        return cls(
            upper_left=bbox.upper_left.copy(),
            bottom_right=bbox.bottom_right.copy(),
            label=label,
            conf=conf
        )

    def draw(self, image: np.ndarray, color: Tuple[int, int, int] = (0, 0, 255), thickness: int = 2) -> np.ndarray:
        # Draw bbox
        super().draw(image, color=color, thickness=thickness)

        # Draw bbox annotation
        y1, x1, _, _ = self.scaled_xyxy_from_image(image)
        annot = f'[{self.label}] {self.conf_annot}'
        # noinspection PyUnresolvedReferences
        image = cv2.putText(image, annot, (y1 + 2, x1 - 4),
                            cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 2)
        # noinspection PyUnresolvedReferences
        image = cv2.putText(image, annot, (y1 + 2, x1 - 4),
                            cv2.FONT_HERSHEY_SIMPLEX, 0.5, color, 1)
        return image



"""
Implementation of ByteTrack.
Reference: https://arxiv.org/abs/2110.06864
"""
from typing import Optional, Dict, Any, List

import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.matching import association_factory
from motrack.tracker.trackers.algorithms.motion_reid import MotionReIDBasedTracker
from motrack.tracker.trackers.catalog import TRACKER_CATALOG
from motrack.tracker.trackers.utils import remove_duplicates
from motrack.tracker.tracklet import Tracklet, TrackletState
from motrack.utils.collections import unpack_n


@TRACKER_CATALOG.register('byte')
class ByteTracker(MotionReIDBasedTracker):
    """
    ByteTrack algorithm.

    Steps:
        0. Estimate all tracklets priors
        1. Split detections into high and low
        2. Match high detections with tracklets with states ACTIVE and LOST using HighMatchAlgorithm
        3. Match remaining ACTIVE tracklets with low detections using LowMatchAlgorithm
        4. Mark remaining (unmatched) tracklets as lost (performed with step 8)
        5. Match NEW tracklets with high detections using NewMatchAlgorithm
            - remove all NEW unmatched tracklets
        6. Initialize new tracklets from unmatched high detections
        7. Update matched tracklets
        8. Delete new unmatched and long-lost tracklets
        9. Delete duplicate between ACTIVE and LOST tracklets
    """
    def __init__(
        self,
        filter_name: str = 'bot-sort',
        filter_params: Optional[dict] = None,
        cmc_name: Optional[str] = None,
        cmc_params: Optional[dict] = None,
        reid_name: Optional[str] = None,
        reid_params: Optional[str] = None,
        high_matcher_algorithm: str = 'default',
        high_matcher_params: Optional[Dict[str, Any]] = None,
        low_matcher_algorithm: str = 'default',
        low_matcher_params: Optional[Dict[str, Any]] = None,
        new_matcher_algorithm: str = 'default',
        new_matcher_params: Optional[Dict[str, Any]] = None,
        detection_threshold: float = 0.6,
        remember_threshold: int = 1,
        initialization_threshold: int = 3,
        new_tracklet_detection_threshold: Optional[float] = None,
        duplicate_iou_threshold: float = 1.00,
        use_observation_if_lost: bool = False,
        appearance_ema_momentum: float = 0.95,
        appearance_buffer: int = 0,
        use_reid_for_low_matching: bool = False
    ):
        if filter_params is None:
            filter_params = {}

        new_tracklet_detection_threshold = new_tracklet_detection_threshold if new_tracklet_detection_threshold is not None \
            else detection_threshold
        super().__init__(
            filter_name=filter_name,
            filter_params=filter_params,
            cmc_name=cmc_name,
            cmc_params=cmc_params,
            reid_name=reid_name,
            reid_params=reid_params,

            new_tracklet_detection_threshold=new_tracklet_detection_threshold,
            remember_threshold=remember_threshold,
            initialization_threshold=initialization_threshold,
            use_observation_if_lost=use_observation_if_lost,

            appearance_ema_momentum=appearance_ema_momentum,
            appearance_buffer=appearance_buffer,
            reid_detection_threshold=detection_threshold if not use_reid_for_low_matching else None
        )

        if high_matcher_algorithm == 'default':
            assert high_matcher_params is None
            high_matcher_algorithm = 'iou'
            high_matcher_params = {
                'match_threshold': 0.2,
                'fuse_score': True
            }

        self._high_match = association_factory(high_matcher_algorithm, high_matcher_params)

        if low_matcher_algorithm == 'default':
            assert low_matcher_params is None
            low_matcher_algorithm = 'iou'
            low_matcher_params = {
                'match_threshold': 0.5
            }

        self._low_match = association_factory(low_matcher_algorithm, low_matcher_params)

        if new_matcher_algorithm == 'default':
            assert new_matcher_params is None
            new_matcher_algorithm = 'iou'
            new_matcher_params = {
                'match_threshold': 0.3,
                'fuse_score': True
            }

        self._new_match = association_factory(new_matcher_algorithm, new_matcher_params)

        # Parameters
        self._detection_threshold = detection_threshold
        self._duplicate_iou_threshold = duplicate_iou_threshold
        self._use_reid_for_low_matching = use_reid_for_low_matching

    def _track(
        self,
        tracklets: List[Tracklet],
        prior_tracklet_bboxes: List[PredBBox],
        detections: List[PredBBox],
        frame_index: int,
        objects_features: Optional[np.ndarray] = None,
        frame: Optional[np.ndarray] = None
    ) -> List[Tracklet]:
        # (1) Split detections into low and high
        high_detections = [d for d in detections if d.conf >= self._detection_threshold]
        high_det_indices = [i for i, d in enumerate(detections) if d.conf >= self._detection_threshold]
        low_detections = [d for d in detections if d.conf < self._detection_threshold]
        low_det_indices = [i for i, d in enumerate(detections) if d.conf < self._detection_threshold]

        # (2) Match high detections with tracklets with states ACTIVE and LOST using HighMatchAlgorithm
        tracklets_active_and_lost_indices, tracklets_active_and_lost, tracklets_active_and_lost_bboxes = \
            unpack_n([(i, t, t_bbox) for i, (t, t_bbox) in enumerate(zip(tracklets, prior_tracklet_bboxes)) if t.is_tracked], n=3)
        high_matches, remaining_tracklet_indices, high_unmatched_detections_indices = \
            self._high_match(tracklets_active_and_lost_bboxes, high_detections,
                             object_features=objects_features[high_det_indices] if self._use_reid_for_low_matching else objects_features,
                             tracklets=tracklets_active_and_lost)
        high_matches = [(tracklets_active_and_lost_indices[t_i], high_det_indices[d_i]) for t_i, d_i in high_matches]
        high_unmatched_detections_indices = [high_det_indices[d_i] for d_i in high_unmatched_detections_indices]
        remaining_tracklet_bboxes = [tracklets_active_and_lost_bboxes[t_i] for t_i in remaining_tracklet_indices]
        remaining_tracklets = [tracklets_active_and_lost[t_i] for t_i in remaining_tracklet_indices]
        remaining_tracklet_indices = [tracklets_active_and_lost_indices[t_i] for t_i in remaining_tracklet_indices]

        # (3) Match remaining ACTIVE tracklets with low detections using LowMatchAlgorithm
        remaining_active_tracklet_indices, remaining_active_tracklets, remaining_active_tracklet_bboxes = \
            unpack_n([(i, t, t_bbox) for i, t, t_bbox in zip(remaining_tracklet_indices, remaining_tracklets, remaining_tracklet_bboxes)
                  if t.state == TrackletState.ACTIVE], n=3)
        remaining_lost_tracklet_indices = \
            [i for i, t in enumerate(tracklets) if t.state == TrackletState.LOST and i in remaining_tracklet_indices]

        low_matches, low_unmatched_tracklet_indices, _ = \
            self._low_match(remaining_active_tracklet_bboxes, low_detections,
                            object_features=objects_features[low_det_indices] if self._use_reid_for_low_matching else None,
                            tracklets=remaining_active_tracklets)
        low_matches = [(remaining_active_tracklet_indices[t_i], low_det_indices[d_i]) for t_i, d_i in low_matches]
        unmatched_tracklet_indices = [remaining_active_tracklet_indices[t_i] for t_i in low_unmatched_tracklet_indices] + \
                                     remaining_lost_tracklet_indices

        # (5) Match NEW tracklets with high detections using NewMatchAlgorithm
        remaining_high_detections = [detections[d_i] for d_i in high_unmatched_detections_indices]
        remaining_high_detection_indices = high_unmatched_detections_indices
        tracklets_new_indices, tracklets_new, tracklets_new_bboxes = \
            unpack_n([(i, t, t_bbox) for i, (t, t_bbox) in enumerate(zip(tracklets, prior_tracklet_bboxes)) if t.state == TrackletState.NEW], n=3)
        new_matches, new_unmatched_tracklets_indices, new_unmatched_detections_indices = \
            self._new_match(tracklets_new_bboxes, remaining_high_detections,
                            object_features=objects_features[remaining_high_detection_indices],
                            tracklets=tracklets_new)
        new_matches = [(tracklets_new_indices[t_i], high_unmatched_detections_indices[d_i]) for t_i, d_i in new_matches]
        new_unmatched_tracklets_indices = [tracklets_new_indices[t_i] for t_i in new_unmatched_tracklets_indices]
        new_unmatched_detections_indices = [remaining_high_detection_indices[d_i] for d_i in new_unmatched_detections_indices]

        # (6) Initialize new tracklets from unmatched high detections
        new_tracklets = self._create_new_tracklets(
            detections=[detections[d_i] for d_i in new_unmatched_detections_indices],
            frame_index=frame_index,
            objects_features=objects_features[new_unmatched_detections_indices] if objects_features is not None else None
        )

        # (7) Update matched tracklets
        all_matches = high_matches + low_matches + new_matches
        matched_detection_indices = [d_i for _, d_i in all_matches]
        matched_objects_features: Optional[List[Optional[np.ndarray]]] = None
        if objects_features is not None:
            # Object's that have features (because of the high detection score)
            # have stored features in the list, otherwise None is set
            detection_to_object_feature_index = {d_i: i for i, d_i in enumerate(high_det_indices)}
            matched_features_indices = [detection_to_object_feature_index.get(d_i) for d_i in matched_detection_indices]
            matched_objects_features = [(objects_features[i] if i is not None else None) for i in matched_features_indices]

        self._update_tracklets(
            matched_tracklets=[tracklets[t_i] for t_i, _ in all_matches],
            matched_detections=[detections[d_i] for d_i in matched_detection_indices],
            matched_object_features=matched_objects_features,
            frame_index=frame_index
        )

        # (8) Delete new unmatched and long-lost tracklets
        self._handle_lost_tracklets(
            lost_tracklets=[tracklets[t_i] for t_i in unmatched_tracklet_indices + new_unmatched_tracklets_indices]
        )

        # (9) Delete duplicate between ACTIVE and LOST tracklets
        deleted_tracklets = [t for t in tracklets if t.state == TrackletState.DELETED]
        active_tracklets = [t for t in tracklets if t.state == TrackletState.ACTIVE]
        lost_tracklets = [t for t in tracklets if t.state == TrackletState.LOST]
        new_tracklets.extend([t for t in tracklets if t.state == TrackletState.NEW])  # Add new tracklets from previous iteration
        active_tracklets, lost_tracklets = remove_duplicates(self._duplicate_iou_threshold, active_tracklets, lost_tracklets)
        return active_tracklets + lost_tracklets + new_tracklets + deleted_tracklets



"""
Tracker association algorithms interface.
"""
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.algorithms.dcm import DCMIoU, MoveDCM
from motrack.tracker.matching.algorithms.iou import IoUAssociation
from motrack.tracker.matching.algorithms.move import Move
from motrack.tracker.matching.algorithms.reid import ReIDIoUAssociation
from motrack.tracker.matching.factory import association_factory



"""
Association methods utilities. Functions:
- Hungarian algorithm
"""
from typing import Tuple, List

import numpy as np
import scipy

INF = 999_999


def hungarian(cost_matrix: np.ndarray) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:
    """
    Performs Hungarian algorithm on arbitrary `cost_matrix`.

    Args:
        cost_matrix: Any cost matrix

    Returns:
        - List of matched (tracklet, detection) index pairs
        - List of unmatched tracklets indices
        - List of unmatched detection indices
    """
    n_tracklets, n_detections = cost_matrix.shape

    # Augment cost matrix with fictional values so that the linear_sum_assignment always has some solution
    augmentation = INF * np.ones(shape=(n_tracklets, n_tracklets), dtype=np.float32) - np.eye(n_tracklets, dtype=np.float32)
    augmented_cost_matrix = np.hstack([cost_matrix, augmentation])

    row_indices, col_indices = scipy.optimize.linear_sum_assignment(augmented_cost_matrix)

    # All row indices that have values above are matched with augmented part of the matrix hence they are not matched
    matches = []
    matched_tracklets = set()
    matched_detections = set()
    for r_i, c_i in zip(row_indices, col_indices):
        if c_i >= n_detections:
            continue  # augmented match -> no match
        matches.append((r_i, c_i))
        matched_tracklets.add(r_i)
        matched_detections.add(c_i)

    unmatched_tracklets = list(set(range(n_tracklets)) - matched_tracklets)
    unmatched_detections = list(set(range(n_detections)) - matched_detections)

    return matches, unmatched_tracklets, unmatched_detections


def greedy(cost_matrix: np.ndarray) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:
    """
    Performs Greedy algorithm on arbitrary `cost_matrix`.
    In general case this does not give better solution than the Hungarian algorithm
    but instead performs association faster.

    Args:
        cost_matrix: Any cost matrix

    Returns:
        - List of matched (tracklet, detection) index pairs
        - List of unmatched tracklets indices
        - List of unmatched detection indices
    """
    n_tracklets, n_detections = cost_matrix.shape
    matches = []
    matched_tracklets = set()
    matched_detections = set()

    for t_i in range(n_tracklets):
        best_match, best_score = None, None
        for d_i in range(n_detections):
            score = cost_matrix[t_i, d_i]
            if score == np.inf or d_i in matched_detections:
                continue

            if best_score is None or score < best_score:
                best_match = d_i
                best_score = score

        if best_match is not None:
            matches.append((t_i, best_match))
            matched_tracklets.add(t_i)
            matched_detections.add(best_match)

    unmatched_tracklets = list(set(range(n_tracklets)) - matched_tracklets)
    unmatched_detections = list(set(range(n_detections)) - matched_detections)
    return matches, unmatched_tracklets, unmatched_detections



"""
Visualization tools for:
- Tracker inference
- Object detection inference
"""
import os
import re
from collections import Counter
from typing import Optional

import cv2
import numpy as np
from tqdm import tqdm

from motrack.config_parser import TrackerVisualizeConfig
from motrack.datasets import BaseDataset
from motrack.evaluation.io import TrackerInferenceReader
from motrack.library.cv import color_palette
from motrack.library.cv.bbox import PredBBox, BBox
from motrack.library.cv.drawing import draw_text
from motrack.library.cv.video_writer import MP4Writer
from motrack.object_detection import DetectionManager


def draw_tracklet(
    frame: np.ndarray,
    tracklet_id: str,
    tracklet_age: int,
    bbox: PredBBox,
    new: bool = False,
    active: bool = True
) -> np.ndarray:
    """
    Draw tracklet on the frame.

    Args:
        frame: Frame
        tracklet_id: Tracklet id
        tracklet_age: How long does the tracklet exist
        bbox: BBox
        new: Is the tracking object new (new id)
            - New object has thick bbox
        active: Is tracklet active or not

    Returns:
        Frame with drawn tracklet info
    """
    color = color_palette.ALL_COLORS_EXPECT_BLACK[int(tracklet_id) % len(color_palette.ALL_COLORS_EXPECT_BLACK)] \
        if active else color_palette.BLACK
    thickness = 5 if new else 2
    frame = BBox.draw(bbox, frame, color=color, thickness=thickness)
    left, top, _, _ = bbox.scaled_xyxy_from_image(frame)
    text = [f'id={tracklet_id}', f'age={tracklet_age}', f'conf={100 * bbox.conf:.0f}%']
    return draw_text(frame, text, left, top, color=color)


def run_visualize_tracker_inference(
    dataset: BaseDataset,
    tracker_active_output: str,
    tracker_output_option: str,
    visualize_cfg: Optional[TrackerVisualizeConfig] = None,
    scene_pattern: str = '(.*?)'
) -> None:
    """
    Visualizes tracker inference output.

    Args:
        dataset: Dataset
        tracker_active_output: Path where tracker ACTIVE tracks inference can be found
        tracker_output_option: Path to any type of tracker inference to visualize (active, all, or postprocess)
            Recommended: All
        visualize_cfg: Custom visualization config (if not set then default is used)
        scene_pattern: Scene pattern filter (regex)

    Returns:

    """
    if visualize_cfg is None:
        visualize_cfg = TrackerVisualizeConfig()

    scene_names = dataset.scenes
    scene_names = [scene_name for scene_name in scene_names if re.match(scene_pattern, scene_name)]
    for scene_name in tqdm(scene_names, desc='Visualizing tracker', unit='scene'):
        scene_info = dataset.get_scene_info(scene_name)
        scene_length = scene_info.seqlength
        imheight = scene_info.imheight
        imwidth = scene_info.imwidth

        active_bboxes = set()
        with TrackerInferenceReader(tracker_active_output, scene_name, image_height=imheight, image_width=imwidth) as tracker_inf_reader:
            for _ in tqdm(range(scene_length), desc=f'Preprocessing "{scene_name}"', unit='frame'):
                read = tracker_inf_reader.read()
                if read is None:
                    continue

                for tracklet_id in read.objects.keys():
                    active_bboxes.add((tracklet_id, read.frame_index))

        scene_video_path = os.path.join(tracker_output_option, f'{scene_name}.mp4')
        with TrackerInferenceReader(tracker_output_option, scene_name, image_height=imheight, image_width=imwidth) as tracker_inf_reader, \
                MP4Writer(scene_video_path, fps=visualize_cfg.fps) as mp4_writer:
            last_read = tracker_inf_reader.read()
            tracklet_presence_counter = Counter()  # Used to visualize new, appearing tracklets

            for index in tqdm(range(scene_length), desc=f'Visualizing "{scene_name}"', unit='frame'):
                image_path = dataset.get_scene_image_path(scene_name, index)
                frame = cv2.imread(image_path)
                assert frame is not None, \
                    f'Failed to load image for frame {index} on scene "{scene_name}" with path "{image_path}"!'

                if last_read is not None and index == last_read.frame_index:
                    for tracklet_id, bbox in last_read.objects.items():
                        tracklet_presence_counter[tracklet_id] += 1
                        draw_tracklet(
                            frame=frame,
                            tracklet_id=tracklet_id,
                            tracklet_age=tracklet_presence_counter[tracklet_id],
                            bbox=bbox,
                            new=tracklet_presence_counter[tracklet_id] <= visualize_cfg.new_object_length,
                            active=(tracklet_id, index) in active_bboxes
                        )

                    last_read = tracker_inf_reader.read()

                mp4_writer.write(frame)

def run_visualize_detections(
    dataset: BaseDataset,
    detection_manager: DetectionManager,
    output_path: str,
    fps: int = 30
) -> None:
    """
    Performs visualization of raw detections without usage of the tracker.

    Args:
        dataset: Dataset to perform tracker inference on
        detection_manager: Detection manager
        output_path: Path where the detections will be stored
        fps: output video fps
    """
    scene_names = dataset.scenes
    for scene_name in tqdm(scene_names, desc='Visualize detections', unit='scene'):
        scene_info = dataset.get_scene_info(scene_name)
        scene_length = scene_info.seqlength
        video_path = os.path.join(output_path, f'{scene_name}.mp4')

        with MP4Writer(video_path, fps=fps) as video_writer:
            for index in tqdm(range(scene_length), desc=f'Visualizing "{scene_name}"', unit='frame'):
                image = dataset.load_scene_image_by_frame_index(scene_name, index)
                detection_bboxes = detection_manager.predict(scene_name, index)
                for bbox in detection_bboxes:
                    image = bbox.draw(image)

                video_writer.write(image)



"""
Support for YOLOX inference (only pretrained models)
"""
from typing import Tuple, Dict, Any, Optional

import numpy as np
import torch

from motrack.object_detection.yolox.yolox_x import DEFAULT_EXP_PATH, DEFAULT_EXP_NAME


class YOLOXPredictor:
    """
    YOLOX predictor (inference) used for already trained model (loaded from checkpoint).
    """
    def __init__(
        self,
        checkpoint_path: str,
        accelerator: str,
        legacy: bool = True,
        conf_threshold: Optional[float] = None,
        exp_path: str = DEFAULT_EXP_PATH,
        exp_name: str = DEFAULT_EXP_NAME
    ):
        """
        Args:
            checkpoint_path: Checkpoint of the pretrained model.
            accelerator: Accelerator
            legacy: Optionally use legacy preprocess
            exp_path: Experiment path (Path to the python script that defines experiments)
            exp_name: Experiment name (Name of the class in the python script that defines experiments)
        """
        from yolox.data.data_augment import ValTransform
        from yolox.exp import get_exp

        self._accelerator = accelerator

        # Load Exp
        self._exp = get_exp(exp_path, exp_name)
        self._conf_threshold = self._exp.test_conf if conf_threshold is None else conf_threshold

        # Load model
        ckpt = torch.load(checkpoint_path)

        model = self._exp.get_model()
        model.load_state_dict(ckpt['model'])
        model.to(self._accelerator)
        model.eval()
        self._model = model

        # Setup preprocess
        self._preprocess = ValTransform(legacy=legacy)

    @torch.no_grad()
    def predict(self, img: np.ndarray) -> Tuple[np.ndarray, Dict[str, Any]]:
        """
        Performs inference on a simple image.

        Args:
            img: Raw image

        Returns:
            - Image bboxes in format: (x1, y1, x2, y1, obj_conf * class_conf, class_conf, class_pred)
            - Image info (used for visualization)
        """
        from yolox.utils import postprocess

        img = img.copy()

        # Store image info
        img_info = {'id': 0}
        height, width = img.shape[:2]
        img_info['height'] = height
        img_info['width'] = width
        img_info['raw_img'] = img

        ratio = min(self._exp.test_size[0] / img.shape[0], self._exp.test_size[1] / img.shape[1])
        img_info['ratio'] = ratio

        # Preprocess image
        img, _ = self._preprocess(img, None, self._exp.test_size)
        img = torch.from_numpy(img).unsqueeze(0).float()
        img = img.to(self._accelerator)

        # Inference
        outputs = self._model(img)
        outputs = outputs.detach().cpu()

        # Postprocess
        outputs = postprocess(outputs, self._exp.num_classes, self._conf_threshold, self._exp.nmsthre, class_agnostic=True)
        output = outputs[0]
        if output is None:
            # Nothing detected
            return np.empty(shape=(0, 7), dtype=np.float32), img_info

        output = output.numpy()
        output[:, 0:4] /= ratio
        output[:, 4] *= output[:, 5]

        # Output format : (x1, y1, x2, y2, obj_conf, class_conf, class_label)
        return output, img_info

    def visualize(self, img: np.ndarray, output: np.ndarray, cls_conf: float = 0.10) -> np.ndarray:
        """
        Visualizes model output on the same image.

        Args:
            img: Same image as in the inference
            output: Model outputs for the same image
            cls_conf: Class confidence threshold

        Returns:
            Image with visualizations
        """
        from yolox.utils import vis

        bboxes = output[:, 0:4]
        cls = output[:, 6]
        scores = output[:, 4]
        vis_res = vis(img, bboxes, scores, cls, cls_conf, self._exp.class_names)

        return vis_res



"""
Utility tracker functions.
"""
from typing import List, Tuple

from motrack.tracker.tracklet import Tracklet


def remove_duplicates(threshold: float, tracklets_lhs: List[Tracklet], tracklets_rhs: List[Tracklet]) -> Tuple[List[Tracklet], List[Tracklet]]:
    """
    Removes duplicates from two list of trackelts, where older tracklet is always kept.
    Two tracklets are considered duplicates if they mostly overlap (defined by the threshold).

    Args:
        threshold: IoU threshold
        tracklets_lhs: First tracklet list
        tracklets_rhs: Second tracklet list

    Returns:
        Updated first and second tracklet lists
    """
    lhs_exclude: List[int] = []
    rhs_exclude: List[int] = []

    for i_lhs, t_lhs in enumerate(tracklets_lhs):
        for i_rhs, t_rhs in enumerate(tracklets_rhs):
            iou_score = t_lhs.bbox.iou(t_rhs.bbox)
            if iou_score >= threshold:
                if t_lhs.age >= t_rhs.age:
                    rhs_exclude.append(i_rhs)
                else:
                    lhs_exclude.append(i_lhs)

    tracklets_lhs = [t_lhs for i_lhs, t_lhs in enumerate(tracklets_lhs) if i_lhs not in lhs_exclude]
    tracklets_rhs = [t_rhs for i_rhs, t_rhs in enumerate(tracklets_rhs) if i_rhs not in rhs_exclude]

    return tracklets_lhs, tracklets_rhs



"""
Visualize detections. Can be used to visually evaluate detector performance without tracking component.
"""
import logging
import os

import hydra
from omegaconf import DictConfig

from motrack.common.project import CONFIGS_PATH
from motrack.config_parser import GlobalConfig
from motrack.datasets import dataset_factory
from motrack.object_detection import DetectionManager
from motrack.tools import run_visualize_detections
from motrack.utils import pipeline

logger = logging.getLogger('VisualizeDetections')


@pipeline.task('visualize-detections')
def inference(cfg: GlobalConfig) -> None:
    output_path = os.path.join(cfg.experiment_path, 'visualize_detections')
    logger.info(f'Saving detection visualizations at "{output_path}".')

    dataset = dataset_factory(
        dataset_type=cfg.dataset.type,
        path=cfg.dataset.fullpath,
        params=cfg.dataset.params,
        test=cfg.eval.split == 'test'
    )

    detection_manager = DetectionManager(
        inference_name=cfg.object_detection.type,
        inference_params=cfg.object_detection.params,
        lookup=cfg.object_detection.load_lookup() if cfg.object_detection.lookup_path is not None else None,
        dataset=dataset,
        cache_path=cfg.object_detection.cache_path,
        oracle=cfg.object_detection.oracle
    )

    run_visualize_detections(
        dataset=dataset,
        detection_manager=detection_manager,
        output_path=output_path,
        fps=cfg.visualize.fps
    )



@hydra.main(config_path=CONFIGS_PATH, config_name='movesort', version_base='1.1')
def main(cfg: DictConfig):
    # noinspection PyTypeChecker
    inference(cfg)


if __name__ == '__main__':
    main()



"""
Interface for composing association algorithms with weights.

Notes:
    - Does not support cascaded algorithm;
    - Each algorithm needs to implement the `_form_cost_matrix`.
"""
from typing import List, Optional

import numpy as np

from motrack.library.cv import PredBBox
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import Tracklet


@ASSOCIATION_CATALOG.register('compose')
class ComposeAssociationAlgorithm(AssociationAlgorithm):
    """
    Allows composition of multiple association matrix.
    """
    def __init__(
        self,
        matchers: List[AssociationAlgorithm],
        weights: List[float],
        fast_linear_assignment: bool = False,
    ):
        """
        Args:
            matchers: List of association algorithms
            weights: Weight of each association algorithm cost matrix
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)

        assert len(matchers) == len(weights), \
            f'Number of matchers and weights must be equal! Got {len(matchers)=} and {len(weights)=}'
        assert all(w >= 0 for w in weights), 'All weights must be non-negative!'
        assert any(w > 0 for w in weights), 'At least one weight must be positive!'

        self._matchers = matchers
        self._weights = weights

    def form_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        weighted_cost_matrix: Optional[np.ndarray] = None
        for matcher, weight in zip(self._matchers, self._weights):
            if weight == 0.0:
                continue

            cost_matrix = weight * matcher.form_cost_matrix(
                tracklet_estimations=tracklet_estimations,
                detections=detections,
                object_features=object_features,
                tracklets=tracklets
            )

            weighted_cost_matrix = cost_matrix if weighted_cost_matrix is None \
                else weighted_cost_matrix + cost_matrix

        return weighted_cost_matrix



"""
Pretty (rich) config print.
"""
import logging
from pathlib import Path
from typing import Sequence

import rich
import rich.syntax
import rich.tree
from omegaconf import DictConfig, OmegaConf

log = logging.getLogger('RichUtils')


def print_config_tree(
    cfg: DictConfig,
    print_order: Sequence[str] = (),
    resolve: bool = False,
    save_to_file: bool = False,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.
    Args:
        cfg (DictConfig): Configuration composed by Hydra.
        print_order (Sequence[str], optional): Determines in what order config components are printed.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
        save_to_file (bool, optional): Whether to export config to the hydra output folder.
    """
    Path(cfg.path.master).mkdir(parents=True, exist_ok=True)

    style = 'dim'
    tree = rich.tree.Tree('CONFIG', style=style, guide_style=style)

    queue = []

    # add fields from `print_order` to queue
    for field in print_order:
        queue.append(field) if field in cfg else log.warning(  # pylint: disable=expression-not-assigned
            f'Field "{field}" not found in config. Skipping "{field}" config printing...'
        )

    # add all the other fields to queue (not specified in `print_order`)
    for field in cfg:
        if field not in queue:
            queue.append(field)

    # generate config tree from queue
    for field in queue:
        branch = tree.add(field, style=style, guide_style=style)

        config_group = cfg[field]
        if isinstance(config_group, DictConfig):
            branch_content = OmegaConf.to_yaml(config_group, resolve=resolve)
        else:
            branch_content = str(config_group)

        branch.add(rich.syntax.Syntax(branch_content, 'yaml'))

    # print config tree
    rich.print(tree)

    # save config tree to file
    if save_to_file:
        with open(Path(cfg.path.master, 'config_tree.log'), 'w', encoding='utf-8') as file:
            rich.print(tree, file=file)



"""
Definition of tracker interface.
"""
from abc import ABC, abstractmethod
from typing import List, Optional
import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.tracklet import Tracklet


class Tracker(ABC):
    """
    Tracker interface.
    """
    def __init__(self):
        self._scene: Optional[str] = None  # Optional for scene specific parameters (etc...)

    def set_scene(self, scene: str) -> None:
        """
        Sets scene value

        Args:
            scene:

        Returns:

        """
        self._scene = scene

    def get_scene(self) -> Optional[str]:
        return self._scene

    @abstractmethod
    def track(
        self,
        tracklets: List[Tracklet],
        detections: List[PredBBox],
        frame_index: int,
        frame: Optional[np.ndarray] = None
    ) -> List[Tracklet]:
        """
        Performs multi-object-tracking step.

        Args:
            tracklets: List of active trackelts
            detections: Lists of new detections
            frame_index: Current frame number
            frame: Pass frame in case CMC or appearance based association is used

        Returns:
            Tracks (active, lost, new, deleted, ...)
        """
        pass

    def reset_state(self) -> None:
        """
        Resets tracker state.
        """
        pass



"""
Height and bottom position distance
"""
from typing import Optional, List

import numpy as np

from motrack.library.cv.bbox import PredBBox
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import Tracklet


@ASSOCIATION_CATALOG.register('hvc')
class HorizontalViewCues(AssociationAlgorithm):
    """
    Heuristic: When camera is positioned horizontally,
        bottom y position (bottom),
        height,
        and bounding box center
        2323become important cues.
    """
    def __init__(
        self,
        bottom_position_factor: float = 1.0,
        height_factor: float = 1.0,
        center_position_factor: float = 0.0,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            bottom_position_factor: Bottom position factor
            height_factor: Height factor
            center_position_factor: Center position factor
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)
        self._bottom_position_factor = bottom_position_factor
        self._height_factor = height_factor
        self._center_position_factor = center_position_factor

    def form_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        _, _ = object_features, tracklets  # Unused

        n_tracklets, n_detections = len(tracklet_estimations), len(detections)
        cost_matrix = np.zeros(shape=(n_tracklets, n_detections), dtype=np.float32)
        for t_i in range(n_tracklets):
            tracklet_bbox = tracklet_estimations[t_i]

            for d_i in range(n_detections):
                det_bbox = detections[d_i]

                bottom_diff = abs(tracklet_bbox.bottom_right.y - det_bbox.bottom_right.y)
                height_diff = abs(tracklet_bbox.height - det_bbox.height)
                center_diff = abs(tracklet_bbox.center.x - det_bbox.center.x)

                cost_matrix[t_i][d_i] = \
                    self._bottom_position_factor * bottom_diff \
                    + self._height_factor * height_diff \
                    + self._center_position_factor * center_diff

        return cost_matrix
